######################################################################################88
from .basic import *
from . import score_trees_devel
from .tcrdist_svg_basic import (
    make_text, SVG_tree_plotter, rgb_from_fraction, create_file)
from . import util
from . import html_colors
from . import tcr_sampler
import scipy.stats
import numpy as np
import pandas as pd

#from functools import reduce
#from mannwhitneyu import mannwhitneyu as mannwhitneyu_exact #too slow

#font_family = 'monospace'
#font_family = 'Droid Sans Mono'
#font_family = 'DejaVu Sans Mono'
font_family = 'courier'

gap_character = '-' ## different from some other places
min_cluster_size = 1


tree_width = 750 #hack=450
ymargin = 30 ## right now we dont want top text to get cut off
xmargin = 10

text_column_separation = 4
labels_tree_separation = 10

branch_width_fraction = 0.1 #hack=0.33333

def _pad_to_middle( s, num ): ## with spaces
    if len(s) >= num:return s
    extra = num-len(s)
    before = extra//2
    after = extra-before
    return ' '*before + s + ' '*after

def _get_primary_number( gene_name ):
    tmp = gene_name[:]
    while tmp and not tmp[0].isdigit():
        tmp = tmp[1:]
    if not tmp:
        ## for example, if gene_name=='TRGJP'
        return 0
    assert tmp[0].isdigit()
    if tmp.isdigit():
        return int(tmp)
    else:
        tmp2 = ''
        while tmp[0].isdigit():
            tmp2 += tmp[0]
            tmp = tmp[1:]
        return int(tmp2)




def make_tall_tcr_tree(
        tcrs,
        organism,
        tcrdist_matrix, # symmetric numpy array
        pngfile,
        tcrdist_clustering_threshold=77.1, # not really important, sets tree order
        color_scores=None,
        color_score_range=None,
        dont_trim_labels=False,
        clone_sizes=None,
        extra_tcr_labels_df=None,
        extra_tcr_labels_to_color_by_color_score=[],
        ignore_nucseqs=False, # if True, don't need cdr3a/b_nucseq in tcrs
        verbose=False,
        cmap=None,
        node_score_func=None, # takes list of leaves, returns node score
):
    ''' tcrs is a list of tuples: tcrs = [ (atcr1,btcr1), (atcr2,btcr2), ....
    atcr1 = (va1, ja1, cdr3a1, cdr3a_nucseq1, *)
    btcr1 = (vb1, jb1, cdr3b1, cdr3b_nucseq1, *)

    minimally tested, currently under development
    '''
    N= len(tcrs)
    assert tcrdist_matrix.shape == (N,N)

    ab = 'AB' # paired trees, for starters

    if color_scores is None:
        color_scores = [1.0]*N
    if color_score_range is None:
        color_score_range = (min(color_scores), max(color_scores))
        if color_score_range[0] == color_score_range[1]:
            color_score_range = (color_score_range[0], color_score_range[0]+.1)
    if clone_sizes is None:
        clone_sizes = [1]*N

    if not ignore_nucseqs:
        if verbose:
            print('parsing junctions for', len(tcrs), 'tcrs')
        junctions = tcr_sampler.parse_tcr_junctions(organism, tcrs)
    else:
        junctions = pd.DataFrame(dict(
            va   =[x[0][0] for x in tcrs],
            ja   =[x[0][1] for x in tcrs],
            cdr3a=[x[0][2] for x in tcrs],
            vb   =[x[1][0] for x in tcrs],
            jb   =[x[1][1] for x in tcrs],
            cdr3b=[x[1][2] for x in tcrs],
        ))

    va_colors, ja_colors, vb_colors, jb_colors = util.assign_colors_to_conga_tcrs(
        tcrs, organism)

    junctions['va_color'] = va_colors
    junctions['ja_color'] = ja_colors
    junctions['vb_color'] = vb_colors
    junctions['jb_color'] = jb_colors

    junctions['color_score'] = color_scores
    junctions['clone_size'] = list(clone_sizes)

    if extra_tcr_labels_df is not None:
        assert extra_tcr_labels_df.shape[0] == N
        extra_tcr_labels = list(extra_tcr_labels_df.columns)
        for label in extra_tcr_labels:
            junctions[label] = list(extra_tcr_labels_df[label])
    else:
        extra_tcr_labels = []



    max_clone_count = np.max(clone_sizes)

    def clonality_fraction( clone_size ):
        #global max_clone_count
        if max_clone_count==1: return 0.0
        exponent = 1.0/3
        mx = max_clone_count**exponent
        cs = clone_size**exponent
        return ( cs-1.0)/(mx-1.0)

    def clonality_color( clone_size ):
        return rgb_from_fraction( clonality_fraction( clone_size ) )


    all_nbrs = []
    for dists in tcrdist_matrix:
        nbrs = []
        for ii,d in enumerate(dists):
            if d <= tcrdist_clustering_threshold:
                nbrs.append( ii )
        all_nbrs.append( nbrs )

    deleted = [False]*N

    centers = []
    all_members = []

    if verbose:
        print(f'clustering {len(tcrs)} tcrs')
    while True:
        clusterno = len(centers)

        best_nbr_count =0
        for i in range(N):
            if deleted[i]: continue
            nbr_count = 0
            for nbr in all_nbrs[i]:
                if not deleted[nbr]:
                    nbr_count+=1
            if nbr_count > best_nbr_count:
                best_nbr_count = nbr_count
                center = i

        if best_nbr_count < min_cluster_size:
            break

        centers.append( center )
        members = [center]

        deleted[center] = True
        for nbr in all_nbrs[center]:
            if not deleted[nbr]:
                deleted[nbr] = True
                members.append( nbr )

        assert len(members) == best_nbr_count
        all_members.append( frozenset(members) )

    num_clusters = len(centers)
    num_tcrs = len(tcrs)

    ## I think this will give a better ordering of the TCRs along the tree
    ## order from 1....N by going through the members of the clusters,
    ##   largest to smallest
    old2new_index = {}
    new2old_index = {}
    last_index=-1
    for members in all_members:
        for member in members:
            last_index += 1
            old2new_index[ member ] = last_index
            new2old_index[ last_index ] = member
    assert len(old2new_index) == num_tcrs

    ## how much vertical space will we need?
    ##
    label_fontsize = 10 #hack=3
    tree_height = label_fontsize * len(tcrs)
    total_svg_height = tree_height + 2*ymargin

    ## now get some info together for plotting
    all_center_dists = {}
    all_scores = []
    sizes = []
    names = []
    infos = [] ## for the color score correlations

    for new_index in range(num_tcrs):
        old_index = new2old_index[ new_index ]
        tcr = junctions.iloc[old_index]
        names.append('')
        sizes.append(1)
        infos.append(f'{tcr.cdr3a[3:-2]} {tcr.cdr3b[3:-2]}')
        all_scores.append( [tcr.color_score] )
        for other_new_index in range(num_tcrs):
            other_old_index = new2old_index[ other_new_index ]
            dist = tcrdist_matrix[ old_index ][ other_old_index ]
            all_center_dists[ (new_index,other_new_index) ] = dist
            all_center_dists[ (other_new_index,new_index) ] = dist

    percentile = -1 # means average

    if node_score_func is None:
        node_scorer = score_trees_devel.CallAverageScore(percentile)
    else:
        node_scorer = (lambda leaves,scores :
                       node_score_func([new2old_index[x] for x in leaves]))

    tree = score_trees_devel.Make_tree_new(
        all_center_dists, len(names), score_trees_devel.Update_distance_matrix_AL,
        all_scores, node_scorer,
    )

    ## look for branches with high/low scores
    #edge_pvals = {}
    #tree_splits_ttest(edge_pvals, tree, [], all_scores, infos,
    #                  epitope+'_'+ab+"_"+color_scheme )


    ## the x-values dont matter here
    ## but the y-values do
    tree_p0 = [10, ymargin ]
    tree_p1 = [1000, tree_height+ymargin ]


    ## node_position tells us where the different clusters are located, vertically
    ##
    tmp_plotter = SVG_tree_plotter()
    node_position,Transform,canvas_tree_min_rmsd, canvas_tree_w_factor \
        = score_trees_devel.Canvas_tree(
            tree, names, sizes, tree_p0, tree_p1, branch_width_fraction,
            tmp_plotter, label_internal_nodes = False,
            score_range_for_coloring = color_score_range )

    cmds = []
    # cmds.append(
    #     make_text(f'#tcrs: {len(tcrs)}', (10,ymargin), 30, font_family=font_family))

    ## now let's add some text for each tcr
    num_columns = 2 + 2*len(ab) + len(extra_tcr_labels)
    text_columns = []
    for i in range( num_columns ):
        text_columns.append( [] )
    header = ['']*num_columns

    #for old_index, tcr in enumerate( tcrs ):
    for old_index, tcr in junctions.iterrows():
        ## 1. cdr3s, indels
        ## 2. clonality text
        ## 3. gene segments text
        ##
        icol = 0

        text0 = ''
        htext = ''
        CDR_START,CDR_STOP = (3,-2) if not dont_trim_labels else (0,None)
        if 'A' in ab:
            if ignore_nucseqs:
                text0 += _pad_to_middle(tcr.cdr3a[CDR_START:CDR_STOP], 15)
                htext += '  cdr3a  '
            else:
                text0 += '{} {} {}'.format(
                    _pad_to_middle(tcr.cdr3a[CDR_START:CDR_STOP], 15),
                    _pad_to_middle(tcr.cdr3a_protseq_masked[CDR_START:CDR_STOP], 15),
                    _pad_to_middle(tcr.a_indels, 6))
                htext += '  cdr3a,cdr3a_masked,a_indels  '
        if 'B' in ab:
            if text0:
                text0 += ' '
            if ignore_nucseqs:
                text0 += _pad_to_middle(tcr.cdr3b[CDR_START:CDR_STOP], 15)
                if htext:
                    htext += '     '
                htext += '    cdr3b'
            else:
                text0 += '{} {} {}'.format(
                    _pad_to_middle(tcr.cdr3b[CDR_START:CDR_STOP], 15),
                    _pad_to_middle(tcr.cdr3b_protseq_masked[CDR_START:CDR_STOP], 15),
                    _pad_to_middle(tcr.b_indels, 6))
                htext += '  cdr3b,cdr3b_masked,b_indels  '

        text_columns[ icol ].append( ( text0, 'black' ) )
        header[icol] = htext
        icol += 1

        ## clonality
        header[icol] = '  C'
        text_columns[icol].append((f'{tcr.clone_size:3d}',
                                   clonality_color(tcr.clone_size)))
        icol += 1

        ## gene segments
        if 'A' in ab:
            text_columns[icol].append(
                (f'{tcr.va[2:4]}{_get_primary_number(tcr.va):02d}', tcr.va_color))
            icol += 1
            text_columns[icol].append(
                (f'{tcr.ja[2:4]}{_get_primary_number(tcr.ja):02d}', tcr.ja_color))
            icol += 1
        if 'B' in ab:
            text_columns[icol].append(
                (f'{tcr.vb[2:4]}{_get_primary_number(tcr.vb):02d}', tcr.vb_color))
            icol += 1
            text_columns[icol].append(
                (f'{tcr.jb[2:4]}{_get_primary_number(tcr.jb):02d}', tcr.jb_color))
            icol += 1

        for label in extra_tcr_labels:
            if label in extra_tcr_labels_to_color_by_color_score:
                mn,mx = color_score_range
                frac = max(0, min(1, (tcr.color_score-mn)/(mx-mn)))
                color = rgb_from_fraction(frac, cmap=cmap)
            else:
                color = 'black'
            text_columns[icol].append((tcr[label], color))
            header[icol] = label
            icol += 1


    ## now go through and figure out how wide each of the text columns is
    x_offset = xmargin
    for col,header_tag in zip( text_columns, header ):
        assert len(col) == num_tcrs
        maxlen = max((len(x[0]) for x in col ))
        if not maxlen: continue
        for old_index, ( text,color ) in enumerate(col):
            new_index = old2new_index[ old_index ]
            ypos = node_position[ new_index ]
            lower_left = [ x_offset, ypos+0.5*label_fontsize*0.75 ]
            cmds.append(
                make_text(text, lower_left, label_fontsize, color=color,
                          font_family=font_family))
        if header_tag:
            max_ypos = max( node_position.values() )
            lower_left = [ x_offset, max_ypos+2.0*label_fontsize*0.75 ]
            htag = header_tag if len(header_tag)<=maxlen else header_tag[:maxlen-1]+'-'
            cmds.append(
                make_text(htag, lower_left, label_fontsize, color='black',
                          font_family=font_family))

        x_offset += text_column_separation + label_fontsize * 0.6 * maxlen


    ## how wide should the tree be?

    tree_p0 = [x_offset + labels_tree_separation, ymargin ]
    tree_p1 = [tree_p0[0] + tree_width, tree_height+ymargin ]

    total_svg_width = tree_p1[0] + xmargin

    ## node_position tells us where the different clusters are located, vertically
    ##

    plotter = SVG_tree_plotter(cmap=cmap)
    node_position,Transform,canvas_tree_min_rmsd, canvas_tree_w_factor = \
        score_trees_devel.Canvas_tree(
            tree, names, sizes, tree_p0, tree_p1, branch_width_fraction,
            plotter, label_internal_nodes = False,
            rmsd_bar_label_stepsize=50, force_min_rmsd=0,
            score_range_for_coloring = color_score_range
        )

    cmds.extend( plotter.cmds )


    ## label pvals
    # if edge_pvals:
    #     plotting_info=(sizes, node_position, Transform,
    #                    canvas_tree_w_factor, canvas_tree_min_rmsd)
    #     label_pval_edges( cmds, edge_pvals, tree, plotting_info )

    create_file(cmds, total_svg_width, total_svg_height, pngfile[:-3]+'svg',
                create_png=True)

    if verbose:
        print('made:', pngfile)
        print('made:', pngfile[:-3]+'svg')



import os
import numpy as np
import pandas as pd
from statsmodels.stats.multitest import fdrcorrection

# from .constants.modules import tcrdist
from .modules.tcrdist.all_genes import all_genes # for recognized genes
from .modules.tcrdist.translation import get_translation
from .modules.tcrdist.tcr_sampler import get_j_cdr3_nucseq
from .preprocessing import IMGT, adaptive_to_imgt_human, adaptive_vfam_mapping

aminoacids = 'ACDEFGHIKLMNPQRSTVWY'
_aminoacids_set = set(aminoacids)

def _is_aaseq(seq:str):
    """
    Check if string contains non-amino acid characters.
    Returns True if string only contains standard amino acid characters.
    """
    try:
        return all(c in _aminoacids_set for c in seq)
    except TypeError:
        return False

def _is_cdr3(seq:str):
    """
    Checks if string is a valid CDR3 amino acid sequence,
    according to the following defenitions:
        - First amino acid character is C.
        - Last amino acid is F or W.
        - Sequence exclusively contains valid amino acid characters.
    """
    try:
        return (_is_aaseq(seq)
            and (seq[0] == 'C')
            and (seq[-1] in ['F', 'W', 'C'])
            and (len(seq) <= 30)
            and (len(seq) >= 4))
    # Exclude non-string type input
    except TypeError:
        return False

def capture_nucseq(tcr, organism='human'):

    vgene, cdr3, rearrangement, jgene = tcr

    # figure out the cdr3 nucseq
    cdr3_nucseq = None
    jgene_cdr3_nucseq = get_j_cdr3_nucseq(organism,jgene).upper() # added .upper()

    for offset in range(3):
        protseq = get_translation( rearrangement, '+{}'.format(offset+1) )
        for ctrim in range(1,4):
            if cdr3[:-ctrim] in protseq:
                start = offset + 3*(protseq.index(cdr3[:-ctrim]))
                length = 3*(len(cdr3)-ctrim)
                cdr3_nucseq = rearrangement[ start : start + length ]
                if ctrim:
                    cdr3_nucseq += jgene_cdr3_nucseq[-3*ctrim:]
                if cdr3 != get_translation( cdr3_nucseq, '+1' ):
                    cdr3_nucseq = ''
                break
        if cdr3_nucseq=='': # failure signal
            cdr3_nucseq = None
            break

    if cdr3_nucseq is None:
        print('parse cdr3_nucseq failed:',tcr)
        return np.nan
    else:
        return cdr3_nucseq

def get_tcr(row):
    return (row.v_call, row.junction_aa, row.junction, row.j_call)

def format_junction(data):
    n_pre = data.shape[0] # number of TCRs in data before parsing
    tcrs = [get_tcr(i) for i in data.itertuples()] # get TCRs in the format (v_call, junction_aa, junction, j_call)
    nt_seq = [capture_nucseq(i) for i in tcrs] # extract the CDR3 nucleotide sequence
    data["junction"] = nt_seq # replace junction column
    data = data.dropna(subset=["junction"]) # drop TCRs with faulty rearrangements
    n_post = data.shape[0] # number of TCRs in data after parsing
    print(f"dropped {n_pre-n_post} TCRs with ambiguous junctions")
    return data

def check_formatting(data : pd.DataFrame, verbose=False):

    # Check if data contains all the necessary columns
    if verbose:
        print("Check if necessary columns are present")
    colnames = data.columns
    required = ["v_call", "j_call", "junction", "junction_aa"]
    for req in required:
        assert req in colnames, f"DataFrame missing column: {req}"
    
    # Check if V/J genes are formatted correctly
    if verbose:
        print("Check V and J gene formatting")
    functional = IMGT[IMGT['fct'].isin(['F','(F)','[F]'])]
    v_check = data.v_call.isin(functional.imgt_allele_name).eq(True).all()
    j_check = data.v_call.isin(functional.imgt_allele_name).eq(True).all()
    assert v_check, "Column 'v_call' formatted improperly\n \
        -- see https://www.imgt.org/IMGTrepertoire/index.php?section=LocusGenes&repertoire=genetable&species=human&group=TRBV (IMGT allele name)for the correct formatting."
    assert j_check, "Column 'j_call' formatted improperly\n \
        -- see https://www.imgt.org/IMGTrepertoire/index.php?section=LocusGenes&repertoire=genetable&species=human&group=TRBJ (IMGT allele name)for the correct formatting."

    # Check formatting of junction
    if verbose:
        print("Check if junction is correctly formatted")
    junction_nt_check = (data.junction.str.len() / data.junction_aa.str.len() == 3).eq(True).all()
    assert junction_nt_check, "CDR3 nucleotide sequence length does not match CDR3 amino acid length\n \
        -- use the 'raptcr.constants.parsing.format_junction' function to correctly parse the junction"
    junction_aa_check = data.junction_aa.apply(lambda x: _is_cdr3(x)).eq(True).all()
    assert junction_aa_check, "Column 'junction_aa' contains non amino-acid characters"

    if verbose:
        print("All checks completed -- DataFrame correctly formatted")


def parse_repertoire(
    file,
    cdr3aa_col="cdr3_b_aa",
    cdr3nt_col="cdr3_b_nucseq",
    vgene_col="v_b_gene",
    jgene_col="j_b_gene",
    sep="\t",
    compression=None,
    drop_cols=None
    ):

    col_remap = {
        cdr3nt_col:"junction",
        cdr3aa_col:"junction_aa",
        vgene_col:"v_call",
        jgene_col:"j_call",
        }
    
    df = pd.read_csv(file, sep=sep, compression=compression)  
    df = df.rename(columns=col_remap)

    # Remove ambiguous CDR3 amino acid sequences
    print("Remove ambiguous CDR3 amino acid sequences")
    df = df[df.junction_aa.apply(lambda cdr3: _is_cdr3(cdr3))]
    
    # Remove ORF and non-functional genes
    print("Parsing V/J genes")
    functional = IMGT[IMGT['fct']=='F']
    df = df[df.v_call.isin(functional.imgt_allele_name)]
    
    # Parse junction (CDR3 nucleotide sequence)
    print("Parsing CDR3 nucleotide sequence")
    df["junction"] = [capture_nucseq((l.v_call, l.junction_aa, l.junction, l.j_call)) for l in df.itertuples()]
    df = df.dropna(subset=["junction_aa","junction","v_call","j_call"])
    
    if drop_cols is not None:
        df = df.drop(columns=drop_cols)
    
    return df

def parse_adaptive_repertoire(name, old=False, recover_unresolved=True):
    # Read file
    df = pd.read_csv(name, sep="\t", low_memory=False)
    # Remove out-of-frame sequences
    df = df[df.frame_type=="In"]
    # Desired columns
    if old:
        prefix = "cdr3_"
    else:
        prefix = ""
    cols = [
        'templates',
        f'{prefix}rearrangement', 
        f'{prefix}amino_acid',
        'v_family',
        'j_family',
        'v_gene',
        'v_allele',
        'j_gene',
        'j_allele'
        ]
    df = df[cols]
    # Gene parsing
    unresolved = df[df.v_gene=="unresolved"]
    df["v_imgt"] = df.v_gene.map(adaptive_to_imgt_human)
    df["j_imgt"] = df.j_gene.map(adaptive_to_imgt_human)
    df = df.dropna(subset=["v_imgt","j_imgt"])
    unresolved["v_imgt"] = unresolved.v_family.map(adaptive_vfam_mapping)
    unresolved["j_imgt"] = unresolved.j_gene.map(adaptive_to_imgt_human)
    if recover_unresolved:
        print(f"Recovered {unresolved.v_imgt.dropna().shape[0]} V genes")
        df = pd.concat([df,unresolved])
    df[["v_allele","j_allele"]] = df[["v_allele","j_allele"]].fillna(0)
    # # df["v_call"] = df.apply(lambda v: v.v_imgt.split("*")[0] + f"*0{int(v.v_allele)}" if v.v_allele!=0 else v, axis=1)
    # # df["j_call"] = df.apply(lambda j: j.j_imgt.split("*")[0] + f"*0{int(j.j_allele)}" if j.j_allele!=0 else j, axis=1)
    df = df.dropna(subset=["v_imgt"])
    print(df.shape)
    df["v_call"] = [df.v_imgt.iloc[n].split("*")[0]+f"*0{int(i)}" if int(i)!=0 else df.v_imgt.iloc[n] for n,i in enumerate(df.v_allele)]
    df["j_call"] = df.j_imgt
    functional = IMGT[IMGT['fct']=='F']
    df = df[df.v_call.isin(functional.imgt_allele_name)]
    df = df[['templates',f'{prefix}rearrangement',f'{prefix}amino_acid','v_call','j_call']]
    df = df[df[f"{prefix}amino_acid"].apply(lambda cdr3: _is_cdr3(cdr3))]
    df = df.dropna(subset=[f'{prefix}rearrangement',f'{prefix}amino_acid','v_call','j_call'])
    df = df.sort_values(by="templates", ascending=False)
    df = df.rename(columns={f"{prefix}rearrangement":"junction",f"{prefix}amino_acid":"junction_aa"})
    df = df.drop_duplicates()
    return df.reset_index(drop=True)


def parse_britanova_repertoires(file, count=2):

    df = pd.read_csv(file, sep='\t', compression='gzip')

    cols = {
        'cdr3nt':'junction', 
        'cdr3aa':'junction_aa', 
        'v':'v_call', 
        'j':'j_call',  
        'count':'clone_count', 
        'freq':'clone_fraction'
        }

    df = df[df['count']>=count]
    df = df[list(cols.keys())]
    df = df.rename(columns=cols)

    # Parse VJ into IMGT format
    df['v_call'] = df['v_call'].apply(lambda x: x.split('(')[0])
    df['v_call'] = df['v_call'].apply(lambda x: x.split('*')[0] + '*01')
    df['j_call'] = df['j_call'].apply(lambda x: x.split('(')[0])
    df['j_call'] = df['j_call'].apply(lambda x: x.split('*')[0] + '*01')

    # Remove ORF and pseudogenes
    functional = IMGT[IMGT['fct']=='F']
    df = df[df.v_call.isin(functional.imgt_allele_name)]
    # Remove ambiguous CDR3 sequences
    df = df[df['junction_aa'].apply(lambda x: _is_cdr3(x))]

    return df.reset_index(drop=True)

def parse_yfv_repertoires(file, output='regular', count=2):
    

    if file.split('.')[-1]=='gz':
        df = pd.read_csv(file, compression='gzip', sep='\t', low_memory=False)
    else:
        df = pd.read_csv(file, sep='\t', low_memory=False)

    # Columns of interest
    if 'AA. Seq. CDR3' in df.columns:
        cols = {
            'AA. Seq. CDR3':'junction_aa', 
            'N. Seq. CDR3':'junction', 
            'All V hits':'v_call', 
            'All J hits':'j_call', 
            'Clone ID':'clonotype_id', 
            'Clone count':'clone_count', 
            'Clone fraction':'clone_fraction'
            }
        # Filter out singlets
        df = df[df['Clone count']>=count]
    else:
        cols = {
            'CDR3.amino.acid.sequence':'junction_aa', 
            'CDR3.nucleotide.sequence':'junction', 
            'bestVGene':'v_call', 
            'bestJGene':'j_call'
            }

    df = df[list(cols.keys())]
    df = df.rename(columns=cols)

    # Parse VJ into IMGT format
    df['v_call'] = df['v_call'].apply(lambda x: x.split('(')[0])
    df['v_call'] = df['v_call'].apply(lambda x: x.split('*')[0] + '*01')
    df['j_call'] = df['j_call'].apply(lambda x: x.split('(')[0])
    df['j_call'] = df['j_call'].apply(lambda x: x.split('*')[0] + '*01')

    # Remove ORF and pseudogenes
    functional = IMGT[IMGT['fct']=='F']
    df = df[df.v_call.isin(functional.imgt_allele_name)]
    # Remove ambiguous CDR3 sequences
    df = df[df['junction_aa'].apply(lambda x: _is_cdr3(x))]

    if output == 'regular':
        return df.reset_index(drop=True)
    elif output == 'alternative':
        cols = {
            'junction_aa':'cdr3aa',
            'junction':'cdr3nt',
            'v_call':'v',
            'j_call':'j',
            'clone_count':'count'
            }
        df = df.rename(columns=cols)
        df = df[list(cols.values())]
        df['cdr3nt'] = df.cdr3nt.str.lower()
        return df.reset_index(drop=True)

def read_sne_file(path, correct_pvals=True, filter_significant=True):
    name = os.path.basename(path).split('_')[0]
    enriched = pd.read_csv(path, sep="\t")
    if correct_pvals:
        enriched["p_adj"] = fdrcorrection(enriched.pval)[1]
    if filter_significant:
        enriched = enriched[enriched.p_adj<.05]
    enriched["clone"] = enriched["v_call"]+"_"+enriched["junction_aa"]
    enriched["sample"] = name
    return enriched



#import os
from scipy.cluster import hierarchy
from scipy.spatial import distance
from .basic import *
from .all_genes import all_genes
from . import html_colors
from collections import Counter
#import parse_tsv

verbose = __name__ == '__main__'

# these helper functions used to do more work, now it's a little silly...

def get_rep( gene, organism ):
    assert gene.startswith('TR')
    return all_genes[organism][gene].rep

def get_mm1_rep( gene, organism ):
    assert gene.startswith('TR')
    return all_genes[organism][gene].mm1_rep

def get_rep_ignoring_allele( gene, organism ):
    rep = get_rep( gene, organism )
    rep = rep[:rep.index('*')]
    return rep

def get_mm1_rep_gene_for_counting( allele, organism ):
    return all_genes[organism][allele].count_rep

def countreps_from_genes( genes, organism ):
    return set( ( all_genes[organism][x].count_rep for x in genes ) )


def tree_sort( old_l, distances, return_leaves=True ): ## average linkage
    assert len(distances) == len(old_l)

    if len(old_l)==1:
        leaves = [0]
    else:
        y = distance.squareform( distances, checks=True )
        Z = hierarchy.average( y )
        #c,coph_dists = hierarchy.cophenet(Z,y)
        leaves = hierarchy.leaves_list( Z )

    new_l = [ old_l[x] for x in leaves ]

    if not return_leaves:
        return new_l
    else:
        return new_l, leaves

def get_top_genes( blast_hits_string ):
    hits = dict( [ ( x.split(':')[0], int( x.split(':')[1] ) ) for x in blast_hits_string.split(';') ] )
    top_score = max( hits.values() )
    return set( [ x for x,y in hits.items() if y >= top_score ] )

def get_top_reps( blast_hits_string, organism ):
    hits = dict( [ ( x.split(':')[0], int( x.split(':')[1] ) ) for x in blast_hits_string.split(';') ] )
    top_score = max( hits.values() )
    return set( [ all_genes[organism][x].rep for x,y in hits.items() if y >= top_score ] )


def reps_from_genes( genes, organism, mm1=False, trim_allele=False ):
    reps = set( ( all_genes[organism][x].mm1_rep for x in genes ) ) if mm1 else \
           set( ( all_genes[organism][x].rep     for x in genes ) )
    if trim_allele:
        reps = set( ( x[:x.index('*')] for x in reps ) )
    return reps

def readme( pngfile, text ):
    """Generate some readme text associated to an image file, that will be incorporated into the
    big html results file by run_basic_analysis.py"""

    out = open(pngfile+'.readme','w')
    cmd = ' '.join(argv)
    out.write("""
<u>Command</u>:
{}
<br><br>
<u>Filename</u>:
{}
<br><br>
<u>Readme</u>:
{}
<br><br>
""".format(cmd, pngfile, text))
    out.close()


## setup a mapping that we can use for counting when allowing mm1s and also ignoring alleles

# allele2mm1_rep_gene_for_counting = {}
# def get_mm1_rep_ignoring_allele( gene, organism ): # helper fxn
#     rep = get_mm1_rep( gene, organism )
#     rep = rep[:rep.index('*')]
#     return rep

# for organism in ['human','mouse']:
#     allele2mm1_rep_gene_for_counting[ organism ] = {}

#     for chain in 'AB':

#         ## look at gene/allele maps
#         vj_alleles = { 'V': [ id for (id,g) in all_genes[organism].iteritems() if g.chain==chain and g.region=='V'],
#                        'J': [ id for (id,g) in all_genes[organism].iteritems() if g.chain==chain and g.region=='J'] }

#         for vj, alleles in vj_alleles.iteritems():
#             gene2rep = {}
#             gene2alleles = {}
#             rep_gene2alleles = {}

#             for allele in alleles:
#                 #assert allele[2] == chain
#                 gene = allele[:allele.index('*')]
#                 rep_gene = get_mm1_rep_ignoring_allele( allele, organism )
#                 if rep_gene not in rep_gene2alleles:
#                     rep_gene2alleles[ rep_gene ] = []
#                 rep_gene2alleles[ rep_gene ].append( allele )

#                 if gene not in gene2rep:
#                     gene2rep[gene] = set()
#                     gene2alleles[gene] = []
#                 gene2rep[ gene ].add( rep_gene )
#                 gene2alleles[gene].append( allele )

#             merge_rep_genes = {}
#             for gene,reps in gene2rep.iteritems():
#                 if len(reps)>1:
#                     assert vj=='V'
#                     if verbose:
#                         print 'multireps:',organism, gene, reps
#                         for allele in gene2alleles[gene]:
#                             print ' '.join(all_genes[organism][allele].cdrs), allele, \
#                                 get_rep(allele,organism), get_mm1_rep(allele,organism)

#                     ## we are going to merge these reps
#                     ## which one should we choose?
#                     l = [ (len(rep_gene2alleles[rep]), rep ) for rep in reps ]
#                     l.sort()
#                     l.reverse()
#                     assert l[0][0] > l[1][0]
#                     toprep = l[0][1]
#                     for (count,rep) in l:
#                         if rep in merge_rep_genes:
#                             assert rep == toprep and merge_rep_genes[rep] == rep
#                         merge_rep_genes[ rep ] = toprep


#             for allele in alleles:
#                 count_rep = get_mm1_rep_ignoring_allele( allele, organism )
#                 if count_rep in merge_rep_genes:
#                     count_rep = merge_rep_genes[ count_rep ]
#                 allele2mm1_rep_gene_for_counting[ organism ][ allele] = count_rep
#                 if verbose:
#                     print 'allele2mm1_rep_gene_for_counting:',organism, allele, count_rep



def assign_label_reps_and_colors_based_on_most_common_genes_in_repertoire( tcr_infos, organism ):
    ## assumes that each element of tcr_infos is a dictionary with fields that would have come from parse_tsv_line
    ## uses the *_countreps info that was filled in by read_pair_seqs.py
    ## the _label_rep* fields get over-written if they were present
    for segtype in segtypes_lowercase:
        countreps_tag = segtype+'_countreps'
        rep_tag       = segtype+'_label_rep'
        color_tag     = segtype+'_label_rep_color' ## where we will store the rep info

        counts = {}
        for tcr_info in tcr_infos:
            reps = tcr_info[countreps_tag].split(';')
            for rep in reps:
                counts[rep] = counts.get(rep,0)+1

        newcounts = {}
        for tcr_info in tcr_infos:
            reps = tcr_info[countreps_tag].split(';')
            toprep = max( [ ( counts[x],x) for x in reps ] )[1]
            tcr_info[rep_tag] = toprep ## doesnt have allele info anymore
            newcounts[toprep] = newcounts.get(toprep,0)+1

        l = [(y,x) for x,y in newcounts.items()]
        l.sort()
        l.reverse()
        rep_colors = dict( list(zip( [x[1] for x in l], html_colors.get_rank_colors_no_lights(len(l)) )) )
        for tcr_info in tcr_infos:
            tcr_info[ color_tag ] = rep_colors[ tcr_info[ rep_tag ] ]

    return ## we modified the elements of the tcr_infos list in place

def assign_colors_to_conga_tcrs( tcrs, organism, return_sorted_color_tuples=False ):
    ''' tcrs is a list [ ( (va,ja,cdr3a,*),(vb,jb,cdr3b,*)), ... ]

    returns a list of 4 lists [ va_colors, ja_colors, vb_colors, jb_colors ] each of length = len(tcrs)

    sorted_color_tuples = [ sorted_va_color_tuples, sorted_ja_color_tuples, .vb., .jb.]
    where sorted_va_color_tuples = [(most_common_va_count_rep, its_color), (second_most_common_va_count_rep, its_color),
                                    ...]

    '''
    color_lists = []
    sorted_color_tuples = []

    for i_ab in range(2):
        for i_vj in range(2):
            genes = [ x[i_ab][i_vj] for x in tcrs ]
            count_reps = [ all_genes[organism][x].count_rep for x in genes ]
            counts = Counter( count_reps )
            uniq_count_reps_sorted = [x[0] for x in counts.most_common()]
            color_tuples = list( zip(uniq_count_reps_sorted,
                                     html_colors.get_rank_colors_no_lights(len(uniq_count_reps_sorted))))
            count_rep_colors = dict(color_tuples)
            color_lists.append( [ count_rep_colors[x] for x in count_reps] )
            sorted_color_tuples.append(color_tuples)
    if return_sorted_color_tuples:
        return color_lists, sorted_color_tuples
    else:
        return color_lists
    #return list(zip(*color_lists))



## this is not exactly perfect, but probably OK to start with...
##
# def detect_fake_chains( clones_file, Achain='A', Bchain='B' ):
#     tcrs = parse_tsv.parse_tsv_file( clones_file, key_fields = [], store_fields = ['va_gene','cdr3a','vb_gene','cdr3b'] )
#     fake_chains = []
#     if len( set( [ (x[0],x[1]) for x in tcrs ] ) )==1:
#         fake_chains.append( Achain )
#     if len( set( [ (x[2],x[3]) for x in tcrs ] ) )==1:
#         fake_chains.append( Bchain )
#     if fake_chains:
#         print('Fake sequence data detected for chains: {}'.format( ' '.join( fake_chains ) ))
#     return fake_chains





# if __name__ == '__main__':
#     for organism in allele2mm1_rep_gene_for_counting:
#         for allele in allele2mm1_rep_gene_for_counting[ organism ]:
#             print 'get_mm1_rep_gene_for_counting



import numpy as np
#import pandas as pd

#from . import basic
from .basic import *
from .amino_acids import amino_acids
#from . import all_genes
from . import tcrdist_svg_basic # hacky code duplication
from . import score_trees_devel
from . import tcr_distances
from . import make_tcr_logo
import random
#from . import util
#from . import tcr_sampler ## for analyze_junction


def make_tcr_tree_svg_commands(
        tcrs,
        organism,
        upper_left,
        width_height,
        tcrdist_matrix, # symmetric numpy matrix
        tcrdist_clustering_threshold=77.1, # try to emulate classic tcr-dist paired threshold (?)
        tcrdist_calculator=None,
        max_tcrs_for_trees=300, # otherwise the actual svg tree gets too tall/dense
        color_scores=None,
        color_score_range=None,
        random_seed=None,
        title=''
):
    ''' tcrs is a list of tuples: tcrs = [ (atcr1,btcr1), (atcr2,btcr2), ....
    atcr1 = (va1, ja1, cdr3a1, cdr3a_nucseq1, *)
    btcr1 = (vb1, jb1, cdr3b1, cdr3b_nucseq1, *)

    upper_left and width_height are tuples (x,y)

    returns the svg cmds

    untested, currently under development
    '''

    assert tcrdist_matrix.shape == (len(tcrs), len(tcrs))
    if color_scores is None:
        color_scores = [1.0]*len(tcrs)
    if color_score_range is None:
        color_score_range = [np.min(color_scores), np.max(color_scores)+1e-6] # add epsilon

    if tcrdist_calculator is None:
        tcrdist_calculator = tcr_distances.TcrDistCalculator(organism)

    font_family = MONOSPACE_FONT_FAMILY # in basic, from ../convert_svg_to_png.py
    gap_character = '-' ## different from some other places
    min_cluster_size = 1
    min_cluster_size_for_glyphs = 5
    min_cluster_fraction_for_glyphs = 0.03

    max_covered_fraction_for_glyphs = 0.75

    max_glyphs = 10

    xmargin = 0.01 * width_height[0]
    ymargin = 0.01 * width_height[1]
    glyphs_to_tree_spacer = 0.01 * width_height[0]
    glyph_size_text_width = 0.01 * width_height[0]

    glyph_region_width = width_height[0] * 1200.0/2200
    ab_glyphs_spacer = 0.03 * glyph_region_width
    tree_width = width_height[0] - glyph_region_width-glyph_size_text_width-glyphs_to_tree_spacer-2*xmargin
    glyph_height = width_height[1] * 135.0/2000
    tree_height = width_height[1] - 2*ymargin

    tree_x0 = upper_left[0] + xmargin + glyph_size_text_width + glyph_region_width + glyphs_to_tree_spacer
    tree_y0 = upper_left[1] + ymargin

    branch_width_fraction = 0.3

    log10_of_zero = -100

    N = len(tcrs)

    all_nbrs = []
    all_dists = tcrdist_matrix #tmphack
    radius = tcrdist_clustering_threshold

    # note that all_nbrs[i] includes i
    for i in range(N):
        all_nbrs.append(np.nonzero(all_dists[i,:] <= radius)[0] )

    deleted = np.full((N,), False)

    centers = []
    all_members = []

    while True:
        clusterno = len(centers)

        nbr_counts = np.array([ np.sum((~deleted)[x]) for x in all_nbrs ]) # all_nbrs is ragged: list of arrays
        nbr_counts[deleted] = 0
        best_nbr_count = np.max(nbr_counts)
        center = np.argmax(nbr_counts)

        if best_nbr_count < min_cluster_size:
            break

        centers.append( center )

        members = [center]
        deleted[center] = True
        for nbr in all_nbrs[center]:
            if not deleted[nbr]:
                deleted[nbr] = True
                members.append( nbr )

        assert len(members) == best_nbr_count
        all_members.append( frozenset(members) )

    if len(centers)<=1:
        # nothing to draw, gives an error below
        print('conga.tcrdist.make_tcr_trees.make_tcr_tree_svg_commands::',
              'WARNING: all tcrs merged into single cluster',
              'no tree to plot', len(tcrs), len(centers))
        return [] # empty list of svg cmds

    ## possibly subsample
    tree_indices = list(range(len(tcrs)))
    if len(tree_indices) > max_tcrs_for_trees:
        rng = random.Random(random_seed) # for reproducibility
        tree_indices = rng.sample( tree_indices, max_tcrs_for_trees )

    ## now get some info together for plotting
    all_center_dists = {}
    all_scores = []
    sizes = []
    names = []

    real_cluster_number2fake_cluster_number = {}
    fake_cluster_number2real_cluster_number = {}
    fake_ic=-1
    real_sizes = []
    for ic,center in enumerate(centers):
        ok_members = []
        for m in all_members[ic]:
            if m in tree_indices:
                ok_members.append(m)
        if not ok_members:
            real_cluster_number2fake_cluster_number[ic] = -1
            continue
        fake_ic += 1
        real_cluster_number2fake_cluster_number[ic] = fake_ic
        fake_cluster_number2real_cluster_number[fake_ic] = ic
        assert fake_ic == len(sizes)
        size = len(ok_members)
        real_size = len(all_members[ic])
        all_scores.append( [ color_scores[x] for x in ok_members ] )
        sizes.append( size )
        real_sizes.append( real_size )
        names.append('')

    ## now do cluster center distances now that we know which ones are in the plot
    for ic,center in enumerate(centers):
        fake_ic = real_cluster_number2fake_cluster_number[ic]
        if fake_ic<0: continue
        for jc,other_center in enumerate(centers):
            fake_jc = real_cluster_number2fake_cluster_number[jc]
            if fake_jc<0: continue
            all_center_dists[ (fake_ic,fake_jc) ] = all_dists[center][other_center]


    #print('num_tcrs:',len(tcrs),'num_clusters:',len(centers),
    #      'fake_num_tcrs',sum(sizes),'fake_num_clusters:',len(sizes))

    cmds = []
    toplabel_fontsize = 15
    toplabel_xpos=upper_left[0]+xmargin
    toplabel_ypos=upper_left[1]+ymargin+0.9*toplabel_fontsize # x,y passed to make_text is the lower_left corner
    if title:
        cmds.append( tcrdist_svg_basic.make_text(title, (toplabel_xpos, toplabel_ypos),
                                                 toplabel_fontsize, font_family=font_family))
        toplabel_ypos += toplabel_fontsize
    cmds.append( tcrdist_svg_basic.make_text('{} xcrs {} clusters'.format(len(tcrs), len(centers)),
                                             (toplabel_xpos, toplabel_ypos),
                                             toplabel_fontsize, font_family=font_family))
    toplabel_ypos += toplabel_fontsize
    cmds.append( tcrdist_svg_basic.make_text('color_range: {:.2f}-{:.2f}'\
                                             .format(color_score_range[0], color_score_range[1]),
                                             (toplabel_xpos, toplabel_ypos),
                                             toplabel_fontsize, font_family=font_family))


    percentile = -1 # I believe this means average the scores when coloring a branch of the tree

    tree = score_trees_devel.Make_tree( all_center_dists, len(names),
                                        score_trees_devel.Update_distance_matrix_AL,
                                        all_scores, percentile )


    plotter = tcrdist_svg_basic.SVG_tree_plotter()

    tree_p0 = [tree_x0, tree_y0]
    tree_p1 = [tree_x0 + tree_width, tree_y0 + tree_height ]


    ## node_position tells us where the different clusters are located, vertically
    ##
    force_min_rmsd = min(74.0, max(0.0, tcrdist_clustering_threshold-3.0))
    node_position, Transform, canvas_tree_min_rmsd, canvas_tree_w_factor = score_trees_devel.Canvas_tree(
        tree, names, sizes, tree_p0, tree_p1, branch_width_fraction,
        plotter, label_internal_nodes = False,
        score_range_for_coloring = color_score_range,
        rmsd_bar_label_stepsize=75, force_min_rmsd=force_min_rmsd )

    max_rmsd_for_glyphs = 3.0*radius

    my_min_cluster_size_for_glyphs = max( min_cluster_size_for_glyphs,
                                          int(floor(0.5 + min_cluster_fraction_for_glyphs*len(tcrs))))

    while True:


        ## get all internal horizontal edges (ie subtrees) that are merged at an rmsd below a threshold
        ## and have at least my_min_cluster_size_for_glyphs
        def get_good_edges( subtree, node_position, sizes, real_sizes ):
            if score_trees_devel.IsALeaf( subtree ):
                return []
            else:
                big_rmsd = subtree[2]
                edges = []
                for ii in range(2):
                    iitree = subtree[ii]
                    little_rmsd = iitree[2]
                    center= score_trees_devel.Center( iitree, node_position, sizes, use_sizes_as_weights=True )
                    real_size = score_trees_devel.Size( iitree, real_sizes )
                    if little_rmsd <= max_rmsd_for_glyphs and real_size >= my_min_cluster_size_for_glyphs:
                        ## this tree is OK
                        edges.append( tuple( ( little_rmsd, big_rmsd, center, real_size,
                                               tuple( sorted( score_trees_devel.Node_members(iitree) ) ) ) ) )
                    edges.extend( get_good_edges( iitree, node_position, sizes, real_sizes ) )
            return edges


        good_edges = get_good_edges( tree, node_position, sizes, real_sizes )

        # if True:
        #     for edge in good_edges:
        #         print 'good:',epitope, ab, edge
            #exit()

        ## figure out which edges we should draw glyphs for
        ## take all the true clusters
        ## try to maximize coverage?
        ##
        glyph_cmds = []
        glyph_edges = []
        while len(glyph_edges) < max_glyphs and len(glyph_edges)<len(good_edges):

            ## figure out who is covered
            covered = set()
            for e in glyph_edges:
                for fake_ic in e[4]:
                    covered.add( fake_ic )

            new_edge = ()
            best_score = -1e6
            for e in good_edges:
                if e in glyph_edges: continue
                (little_rmsd,big_rmsd,center,real_size,clusters) = e
                new_size = real_size
                for fake_ic in clusters:
                    if fake_ic in covered:
                        new_size -= real_sizes[fake_ic]
                if new_size<my_min_cluster_size_for_glyphs: continue
                if (real_size-new_size) > max_covered_fraction_for_glyphs * real_size: continue

                if little_rmsd<1e-3: ## true cluster
                    sortscore = 1000 + new_size
                else:
                    sortscore = new_size - ( little_rmsd / max_rmsd_for_glyphs ) * len(tcrs)
                if sortscore > best_score:
                    new_edge = e[:]
                    best_score = sortscore
            if not new_edge: break
            #print 'new_edge:',new_edge
            glyph_edges.append( new_edge )

            ## let's make a box around this edge in the tree
            (little_rmsd,big_rmsd,center,real_size,clusters) = new_edge
            fake_size = sum( [sizes[x] for x in clusters] )
            line_width = max(1,int(floor(0.5+ fake_size*canvas_tree_w_factor )))
            box_x0 = Transform(max(canvas_tree_min_rmsd,little_rmsd)) ; box_x1 = Transform(big_rmsd)
            assert little_rmsd<big_rmsd
            glyph_cmds.append( tcrdist_svg_basic.rectangle( ( box_x0, center - line_width/2.0 ),
                                                            ( box_x1, center + line_width/2.0 ),
                                                            'none', 'black', stroke_width=3, dashed=True ) )
            glyph_cmds.append( tcrdist_svg_basic.make_text( '%d'%real_size, ( box_x1+4, center + 0.75*15.0/2) ,
                                                            15, font_family=font_family))

        if len( glyph_edges ) < max_glyphs and my_min_cluster_size_for_glyphs>1:
            my_min_cluster_size_for_glyphs -= 1
        else:
            break

    cmds.extend(glyph_cmds)

    min_glyph_loc = toplabel_ypos + glyph_height/2
    max_glyph_loc = tree_p1[1] - glyph_height/2

    #glyph_location = dict( [ (e,max(min_glyph_loc,min(max_glyph_loc,e[2]))) for e in glyph_edges ] )
    glyph_location = dict( [ (e,e[2]) for e in glyph_edges ] )



    while True: ## keep looping until we are bump-free
        l = [(y,x) for x,y in glyph_location.items() ]

        l.sort()
        bump = False
        stepsize = 5 ## pixels
        for (loc1,edge1),(loc2,edge2) in zip( l[:-1], l[1:] ):
            sep = loc2-loc1
            if sep < 1.25*glyph_height:
                bump = True
                if glyph_location[edge1]-stepsize > min_glyph_loc:
                    glyph_location[edge1] -= stepsize
                if glyph_location[edge2]+stepsize < max_glyph_loc:
                    glyph_location[edge2] += stepsize
                #print('bump1', loc1, loc2)
        if bump:
            continue

        # first try to move UP if necessary (ie lower loc)
        # move the highest one first
        # these are sorted from smallest to largest location, ie from highest to lowest since y-location increase down
        for (loc,edge) in l:
            if glyph_location[edge] > max_glyph_loc:
                glyph_location[edge] -= stepsize
                #print('bump2', loc)
                bump = True
        if bump:
            continue

        for (loc,edge) in reversed(l): # from bottom to top
            if glyph_location[edge] < min_glyph_loc:
                glyph_location[edge] += stepsize
                #print('bump3', loc)
                bump = True

        if not bump:
            break

    #exit()

    ## now let's draw some cluster summary glyphs next to the cluster tree


    ## write out the glyph-cluster sizes
    for glyph_edge in glyph_edges:
        real_size = glyph_edge[3]
        loc = glyph_location[glyph_edge]
        ## silly xloc, was hard-coded to 5
        cmds.append( tcrdist_svg_basic.make_text('%3d'%real_size, ( upper_left[0]+xmargin-5, loc+glyph_height/2.) , 15,
                                                 font_family=font_family))



    for ii_ab2, ab2 in enumerate('AB'):
        single_glyph_width = float(glyph_region_width-ab_glyphs_spacer) / 2 # A or B

        for glyph_edge in glyph_edges:
            (little_rmsd, big_rmsd, y_center, real_size, clusters ) = glyph_edge

            members = []
            for fake_ic in clusters:
                members.extend( all_members[ fake_cluster_number2real_cluster_number[ fake_ic ] ] )

            assert len(members) == real_size

            y0 = glyph_location[glyph_edge] - glyph_height/2.
            x0 = upper_left[0] + xmargin + glyph_size_text_width + ii_ab2 * ( single_glyph_width + ab_glyphs_spacer )

            glyph_tcrs = [tcrs[x] for x in members]
            cmds.extend(make_tcr_logo.make_tcr_logo_svg_commands_for_tcrs(
                glyph_tcrs, ab2, organism, [x0,y0], [single_glyph_width, glyph_height], tcrdist_calculator))


    return plotter.cmds + cmds







from .basic import *
from .all_genes import all_genes, gap_character
from .amino_acids import amino_acids
from .tcr_distances_blosum import blosum, bsd4

## see the TcrDistCalculator at the end for simple tcrdist calculations


##############################################33

GAP_PENALTY_V_REGION = 4
GAP_PENALTY_CDR3_REGION = 12 # same as GAP_PENALTY_V_REGION=4 since WEIGHT_CDR3_REGION=3 is not applied
WEIGHT_V_REGION = 1
WEIGHT_CDR3_REGION = 3
DISTANCE_MATRIX = bsd4
ALIGN_CDR3S = False
TRIM_CDR3S = True


def blosum_character_distance( a, b, gap_penalty ):
    if a== gap_character and b == gap_character:
        return 0
    elif a == '*' and b == '*':
        return 0
    elif a == gap_character or b == gap_character or a=='*' or b=='*':
        return gap_penalty
    else:
        # assert a in amino_acids
        # assert b in amino_acids
        # maxval = min( blosum[(a,a)], blosum[(b,b)] )
        # return maxval - blosum[(a,b)]
        return DISTANCE_MATRIX[ (a,b) ]

def blosum_sequence_distance( aseq, bseq, gap_penalty ):
    assert len(aseq) == len(bseq)
    dist = 0.0
    for a,b in zip(aseq,bseq):
        if a == ' ':
            assert b== ' '
        else:
            dist += blosum_character_distance( a, b, gap_penalty )
    return dist

def align_cdr3_regions( a, b, gap_character ):
    if len(a) == len(b):
        return (a[:],b[:])

    if len(a)<len(b): ## s0 is the shorter sequence
        s0,s1 = a,b
    else:
        s0,s1 = b,a

    lendiff = len(s1)-len(s0)

    best_score=-1000
    best_gappos=0 # in case len(s0) == 1

    # the gap comes after s0[gappos]

    for gappos in range(len(s0)-1):
        score=0
        for i in range(gappos+1):
            score += blosum[ (s0[i],s1[i]) ]
        for i in range(gappos+1,len(s0)):
            score += blosum[ (s0[i],s1[i+lendiff]) ]
        if score>best_score:
            best_score = score
            best_gappos = gappos
    ## insert the gap
    s0 = s0[:best_gappos+1] + gap_character*lendiff + s0[best_gappos+1:]

    assert len(s0) == len(s1)

    if len(a)<len(b): ## s0 is the shorter sequence
        return ( s0, s1 )
    else:
        return ( s1, s0 )

## align
##
##   shortseq[        ntrim: gappos   ] with longseq[ ntrim: gappos ] and
##   shortseq[ -1*remainder: -1*ctrim ] with longseq[ -1*remainder: -1*ctrim ]
##
## but be careful about negative indexing if ctrim is 0
##
## the gap comes after position (gappos-1) ie there are gappos amino acids before the gap
##
##
## DOES NOT INCLUDE THE GAP PENALTY
##
def sequence_distance_with_gappos( shortseq, longseq, gappos ):
    ntrim = 3 if TRIM_CDR3S else 0
    ctrim = 2 if TRIM_CDR3S else 0
    remainder = len(shortseq)-gappos
    dist = 0.0
    count =0
    if ntrim < gappos:
        for i in range(ntrim,gappos):
            dist += DISTANCE_MATRIX[ (shortseq[i], longseq[i] ) ]
            count += 1
    if ctrim < remainder:
        for i in range(ctrim, remainder):
            dist += DISTANCE_MATRIX[ (shortseq[-1-i], longseq[-1-i] ) ]
            count += 1
    return dist, count


def weighted_cdr3_distance( seq1, seq2 ):
    shortseq,longseq = (seq1,seq2) if len(seq1)<=len(seq2) else (seq2,seq1)

    ## try different positions of the gap
    lenshort = len(shortseq)
    lenlong = len(longseq)
    lendiff = lenlong - lenshort

#    assert lenshort>3 ##JCC testing
    assert lenshort > 1##JCC testing
    assert lendiff>=0
    if TRIM_CDR3S:
        # used to say >3+2 but I don't think that's necessary
        assert lenshort >= 3+2 ## something to align... NOTE: Minimum length of cdr3 protein carried into clones file is currently set in the read_sanger_data.py script!

    if not ALIGN_CDR3S:
        ## if we are not aligning, use a fixed gap position relative to the start of the CDR3
        ## that reflects the typically longer and more variable-length contributions to
        ## the CDR3 from the J than from the V. For a normal-length
        ## CDR3 this would be after the Cys+5 position (ie, gappos = 6; align 6 rsds on N-terminal side of CDR3).
        ## Use an earlier gappos if lenshort is less than 11.
        ##
        gappos = min( 6, 3 + (lenshort-5)//2 )
        best_dist, count = sequence_distance_with_gappos( shortseq, longseq, gappos )

    else:
        ## the CYS and the first G of the GXG are 'aligned' in the beta sheet
        ## the alignment seems to continue through roughly CYS+4
        ## ie it's hard to see how we could have an 'insertion' within that region
        ## gappos=1 would be a insertion after CYS
        ## gappos=5 would be a insertion after CYS+4 (5 rsds before the gap)
        ## the full cdr3 ends at the position before the first G
        ## so gappos of len(shortseq)-1 would be gap right before the 'G'
        ## shifting this back by 4 would be analogous to what we do on the other strand, ie len(shortseq)-1-4
        min_gappos = 5
        max_gappos = len(shortseq)-1-4
        while min_gappos>max_gappos:
            min_gappos -= 1
            max_gappos += 1
        for gappos in range( min_gappos, max_gappos+1 ):
            dist, count = sequence_distance_with_gappos( shortseq, longseq, gappos )
            if gappos>min_gappos:
                assert count==best_count
            if gappos == min_gappos or dist < best_dist:
                best_dist = dist
                best_gappos = gappos
                best_count = count
        #print 'align1:',shortseq[:best_gappos] + '-'*lendiff + shortseq[best_gappos:], best_gappos, best_dist
        #print 'align2:',longseq, best_gappos, best_dist


    ## Note that WEIGHT_CDR3_REGION is not applied to the gap penalty
    ##
    return  WEIGHT_CDR3_REGION * best_dist + lendiff * GAP_PENALTY_CDR3_REGION

def compute_all_v_region_distances( organism ):
    rep_dists = {}
    for chain in 'AB': # don't compute inter-chain distances
        repseqs = []
        for id,g in all_genes[organism].items():
            if g.chain == chain and g.region == 'V':
                merged_loopseq = ' '.join( g.cdrs[:-1])
                repseqs.append( ( id, merged_loopseq  ) )
                rep_dists[ id ] = {}

        for r1,s1 in repseqs:
            for r2,s2 in repseqs:
                #if r1[2] != r2[2]: continue
                rep_dists[r1][r2] = WEIGHT_V_REGION * \
                                    blosum_sequence_distance( s1, s2, GAP_PENALTY_V_REGION )

    return rep_dists

def compute_distance(t1, t2, chains, rep_dists):
    ''' t1/2 = [ va_reps, vb_reps, l['cdr3a'], l['cdr3b'] ]
    '''
    dist=0.0
    if 'A' in chains:
        dist += min( ( rep_dists[x][y] for x in t1[0] for y in t2[0] ) ) +\
                weighted_cdr3_distance( t1[2], t2[2] )
    if 'B' in chains:
        dist += min( ( rep_dists[x][y] for x in t1[1] for y in t2[1] ) ) +\
                weighted_cdr3_distance( t1[3], t2[3] )
    return dist

##################################################################################################################
##################################################################################################################

class TcrDistCalculator:
    ''' Simple tcr dist calculator
    precomputes the V-region distances at construction time

    use like this:

    tdist = TcrDistCalculator('human')

    d = tdist(tcr1, tcr2)

    tcr1 and tcr2 are both paired tcrs, represented as tuples of tuples:

    tcr1/2 = ( (va, ja, cdr3a,...), (vb, jb, cdr3b,...) )

    what we really need is:
    * tcr1[0][0] = V-alpha gene
    * tcr1[0][2] = CDR3-alpha
    * tcr1[1][0] = V-beta gene
    * tcr1[1][2] = CDR3-beta

    and same for tcr2
    '''
    def __init__(self, organism):
        self.rep_dists = compute_all_v_region_distances(organism)

    def __call__(self, tcr1, tcr2):

        ''' tcr1 and tcr2 are both paired tcrs, represented as tuples of tuples:

        tcr1/2 = ( (va, ja, cdr3a,...), (vb, jb, cdr3b,...) )

        what we really need is:
        * tcr1[0][0] = V-alpha gene
        * tcr1[0][2] = CDR3-alpha
        * tcr1[1][0] = V-beta gene
        * tcr1[1][2] = CDR3-beta

        and same for tcr2
        '''

        return ( self.rep_dists[tcr1[0][0]][tcr2[0][0]] + weighted_cdr3_distance(tcr1[0][2], tcr2[0][2]) +
                 self.rep_dists[tcr1[1][0]][tcr2[1][0]] + weighted_cdr3_distance(tcr1[1][2], tcr2[1][2]) )


    def single_chain_distance(self, chain1, chain2 ):
        ''' chain1 and chain2 are single-chain tcrs, represented as tuples:

        chain1 = (va, ja, cdr3a,...)  OR  chain1 = (vb, jb, cdr3b,...)

        what we really need is:
        * chain1[0] = V-alpha gene   (or V-beta)
        * chain1[2] = CDR3-alpha     (or CDR3-beta)

        and same for chain2
        '''
        return self.rep_dists[chain1[0]][chain2[0]] + weighted_cdr3_distance(chain1[2], chain2[2])







import pandas as pd
from os.path import dirname, abspath, join
from .base import GAPCHAR

ROOT = dirname(dirname(dirname(abspath(__file__))))
DATA = join(ROOT, 'snetcr/constants/data')

IMGT = pd.read_csv(join(DATA,'imgt_reference.tsv'), sep='\t')
mapping = pd.read_csv(join(DATA, 'adaptive_imgt_mapping.csv'))

adaptive_to_imgt_human = mapping.loc[mapping['species'] == 'human'].set_index('adaptive')['imgt'].fillna('NA').to_dict()
adaptive_to_imgt_mouse = mapping.loc[mapping['species'] == 'mouse'].set_index('adaptive')['imgt'].fillna('NA').to_dict()

v_fam_freq = pd.read_csv(join(DATA,"adaptive_v_fam_to_imgt_gene.txt"), sep="\t")
adaptive_vfam_mapping = dict(zip(v_fam_freq.adaptive_v_family, v_fam_freq.imgt_v_allele))

def vgene_to_cdr():
    '''
    Load V gene reference file that contains mapping information to extract 
    CDR1 and CDR2 information based on the V gene allele annotation.
    '''
    return pd.read_csv(join(DATA,'vgene_to_cdr.txt'), sep='\t')

def get_gene_reference():
    return pd.read_csv(join(DATA,'combo_xcr.tsv'), sep='\t')

def add_cdr_columns(df:pd.DataFrame, vcol:str='v_call') -> pd.DataFrame:
    '''
    Adds CDR1 and CDR2 information to TCR sequences based on V gene allele annotation.
    '''
    vgene = vgene_to_cdr()
    df['cdr1_b_aa'] = df[vcol].map(dict(zip(vgene['id'],vgene['cdr1'])))
    df['cdr2_b_aa'] = df[vcol].map(dict(zip(vgene['id'],vgene['cdr2'])))
    return df

def to_tcrdist3_format(df:pd.DataFrame, vgenecol:str='v_call', jgenecol:str='j_call', cdr3col:str='junction_aa'):
    return df.rename(columns={
        vgenecol:'v_b_gene',
        jgenecol:'j_b_gene',
        cdr3col:'cdr3_b_aa'
    })

def format_organism(organism):
    '''
    Correctly format the input organism.
    '''
    mapping = {
        'human':'human',
        'homo':'human',
        'homo_sapiens':'human',
        'homosapiens':'human',
        'sapiens':'human',
        'mouse':'mouse',
        'mus':'mouse',
        'musmusculus':'mouse',
        'mus_musculus':'mouse',
        'musculus':'mouse'
        }
    assert organism.lower() in mapping, f"Unknown organism: {organism}. Please select human or mouse."
    return mapping[organism.lower()]

def format_chain(chain):
    '''
    Correctly format the input TCR chain.
    '''
    mapping = {
        'b':'B',
        'beta':'B',
        'trb':'B',
        'trbeta':'B',
        'tcrb':'B',
        'tcrbeta':'B',
        'tcr_beta':'B',
        'a':'A',
        'alpha':'A',
        'tra':'A',
        'tralpha':'A',
        'tcra':'A',
        'tcralpha':'A',
        'tcr_alpha':'A',
        'ab':'AB',
        'alphabeta':'AB',
        'alpha_beta':'AB',
        'g':'G',
        'gamma':'G',
        'trg':'G',
        'trgamma':'G',
        'tcrg':'G',
        'tcrgamma':'G',
        'tcr_gamma':'G',
        'd':'D',
        'delta':'D',
        'trd':'D',
        'trdelta':'D',
        'tcrd':'D',
        'tcrdelta':'D',
        'tcr_delta':'D',
        'gd':'GD',
        'gammadelta':'GD',
        'gamma_delta':'GD',
    }
    assert chain.lower() in mapping, f"Unknown chain: {chain}. Please select A, B or AB."
    return mapping[chain.lower()]

def setup_gene_cdr_strings(organism:str='human', chain:str='B'):
    ''' 
    Returns dict mapping vgene names to concatenated cdr1-cdr2-cdr2.5 strings
    columns without any sequence variation (e.g. all gaps) are removed
    '''
    # Make sure organism and chain input is correctly formatted
    organism = format_organism(organism)
    chain = format_chain(chain)
    # Get CDR information from gene reference file
    all_genes_df = get_gene_reference()
    all_genes_df = all_genes_df[(all_genes_df.organism==organism)&
                                (all_genes_df.chain.isin(list(chain)))&
                                (all_genes_df.region=='V')]
    assert all_genes_df.id.value_counts().max()==1
    all_genes_df.set_index('id', inplace=True)
    all_genes_df['cdrs'] = all_genes_df.cdrs.str.split(';')
    vgenes = list(all_genes_df.index)
    gene_cdr_strings = {x:'' for x in vgenes}
    oldgap = '.' # gap character in the all_genes dict
    # print(gene_cdr_strings)
    for icdr in range(3):
        cdrs = all_genes_df.cdrs.str.get(icdr).str.replace(oldgap,GAPCHAR,regex=False)
        L = len(cdrs[0])
        for i in reversed(range(L)):
            # print(L)
            col = set(x[i] for x in cdrs)
            if len(col) == 1: # no variation
                cdrs = [x[:i]+x[i+1:] for x in cdrs]
        for g, cdr in zip(vgenes, cdrs):
            gene_cdr_strings[g] += cdr
    return gene_cdr_strings

def detect_vgene_col(df):
    pattern = r'TR[A,B,G,D]V'
    matching_columns = [column for column in df.columns if df[column].astype(str).str.contains(pattern, regex=True).all()]
    if len(matching_columns) == 1:
        print("Autodetected V gene column:", matching_columns[0])
        return matching_columns[0]
    elif len(matching_columns) == 0:
        raise ValueError("No V gene column detected. If your dataframe contains V gene information, please make sure it is IMGT-formatted.")
    else:
        raise ValueError("Multiple V gene columns detected. Please specify the correct column name.")

def detect_cdr3_col(df):
    pattern = r'^[CW].*[FWC]$'
    matching_columns = [column for column in df.columns if df[column].astype(str).str.contains(pattern, regex=True).all()]
    if len(matching_columns) == 1:
        print("Autodetected CDR3AA column:", matching_columns[0])
        return matching_columns[0]
    elif len(matching_columns) == 0:
        raise ValueError("No CDR3 column detected. Please make sure your dataframe contains a column with CDR3 sequences or specify the column name.")
    else:
        raise ValueError("Multiple CDR3 columns detected. Please specify the correct column name.")


import time
from .constants.base import AALPHABET

from collections import Counter
import numpy as np

def timed(myfunc):
    # Decorator to keep track of time required to run a function
    def timed(*args, **kwargs):
        start = time.time()
        result = myfunc(*args, **kwargs)
        end = time.time()
        print(f"Total time to run '{myfunc.__name__}': {(end-start):.3f}s")
        return result

    return timed

def profile_matrix(sequences : list):
    '''
    Calculates the profile matrix for a set of sequences (i.e. all cluster members).
    NOTE: this version does not take into account the expected frequency of each amino acid at each position.
    '''

    # Make sure to proceed only if all sequences in the cluster have equal length
    seq_len = len(sequences[0])
    if not all(len(seq) == seq_len for seq in sequences):

        # On the rare occasion that a cluster contains sequences of inequal length.
        # Typically, there is/are only one (or very few) sequence(s) that differ from the avg. CDR3 length in the cluster.
        # Therefore, we use the length of the highest proportion of sequences as the standard, and delete all others.
        seq_len = Counter([len(s) for s in sequences]).most_common()[0][0]
        sequences = [s for s in sequences if len(s) == seq_len]
    
    # Initiate profile matrix with zeros
    pm = np.zeros(shape=(len(AALPHABET), seq_len))

    # initiate AA dict:
    AAs = {aa: i for i, aa in enumerate(AALPHABET)}

    # Fill in profile matrix with counts
    for s in sequences:
        for i, aa in enumerate(s):
            pm[AAs[aa], i] += 1

    # normalize profile matrix to percentages
    pm = pm / len(sequences)

    return pm


def motif_from_profile(profile, method, cutoff=.7):
    '''
    Generate consensus sequence motif from a profile matrix.
    Square brackets [...] indicate multiple aa possibilities at that position.
    X represents any aa.
    '''
    AA_map = {i:aa for i,aa in enumerate(AALPHABET)}

    consensus = ''
    
    if method.lower() == 'standard':
        top_idxs = np.argpartition(profile, -2, axis=0)[-2:].T
        top_values = np.partition(profile, -2, axis=0)[-2:].T
        for (second_max_idx, max_idx), (second_max_value, max_value) in zip(top_idxs, top_values):
            if max_value >= cutoff:
                consensus += AA_map[max_idx]
            elif max_value + second_max_value >= cutoff:
                if max_value >= 2*second_max_value:
                    consensus += AA_map[max_idx].lower()
                else:
                    consensus += f"[{AA_map[max_idx]}{AA_map[second_max_idx]}]"
            else:
                consensus += "."
                
    elif method.lower() == 'conservative':
        max_idx, max_value = np.argmax(profile.T, axis=1), np.amax(profile.T, axis=1)
        for idx, value in zip(max_idx, max_value):
            if value > cutoff:
                consensus += AA_map[idx]
            else:
                consensus += "."

    return consensus

def dump_json(neighbor_dict, file):
    json_string = json(neighbor_dict)
    f = open(file,"w")
    f.write(json_string)
    f.close()

def load_json(self, file):
    with open(file) as json_file:
        neighbors = json.load(json_file)
    return neighbors


import numpy as np
import pandas as pd
import itertools

from concurrent.futures import ThreadPoolExecutor   
from os.path import dirname, abspath, join
from collections import Counter

from .constants.modules.tcrdist.tcr_sampler import parse_tcr_junctions,  resample_shuffled_tcr_chains
from .constants.preprocessing import format_chain
from .constants.base import AALPHABET
from .pgen import generate_sequences


ROOT = dirname(dirname(dirname(abspath(__file__))))
MODELS = join(ROOT,'immune_response_detection/snetcr/constants/modules/olga/default_models/human_T_beta/')
DATADIR = join(ROOT, 'immune_response_detection/snetcr/constants/data/')

def get_vfam(df, vcol='v_call'):
    '''
    Extract V gene family information from V gene column.
    '''
    return df[vcol].apply(lambda x: x.split('*')[0].split('-')[0])

def get_jfam(df, jcol='j_call'):
    '''
    Extract J gene family information from J gene column.
    '''
    return df[jcol].apply(lambda x: x.split('*')[0].split('-')[0])

def filter_out_bad_genes_and_cdr3s(
        df,
        v_column,
        cdr3_column,
        organism,
        chain,
        min_cdr3_len = 6,
        j_column=None,
):
    ''' returns filtered copy of df

    removes tcrs with

    * unrecognized V gene names (and J genes, if j_column != None)
    * V genes whos cdr1/cdr2/cdr2.5 contain '*' (probably pseudogenes?)
    * CDR3s with non-AA characters or shorter than 6
    '''
    all_genes_df = pd.read_table(DATADIR+'combo_xcr.tsv')
    all_genes_df = all_genes_df[(all_genes_df.organism==organism)&
                                (all_genes_df.chain==chain)]
    # drop cdr3, since a '*' there might be trimmed back so it's OK...
    all_genes_df['cdrs'] = all_genes_df.cdrs.str.split(';').str.slice(0,3).str.join(';')

    known_v_genes = set(all_genes_df[all_genes_df.region=='V'].id)
    bad_v_genes = set(x.id for x in all_genes_df.itertuples()
                      if x.region == 'V' and '*' in x.cdrs)
    print('bad_v_genes:', len(bad_v_genes), bad_v_genes)

    good_cdr3s_mask = np.array(
        [len(cdr3)>=min_cdr3_len and all(aa in AALPHABET for aa in cdr3)
         for cdr3 in df[cdr3_column]])
    print('bad_cdr3s in df:', (~good_cdr3s_mask).sum())

    bad_v_genes_mask = df[v_column].isin(bad_v_genes)
    print('bad_v_genes in df:', bad_v_genes_mask.sum(),
          df[bad_v_genes_mask][v_column].unique())

    unknown_genes_mask = ~df[v_column].isin(known_v_genes)
    print('unknown_genes in df:', unknown_genes_mask.sum(),
          df[unknown_genes_mask][v_column].unique())
    if j_column is not None:
        known_j_genes = set(all_genes_df[all_genes_df.region=='J'].id)
        unknown_j_genes_mask = ~df[j_column].isin(known_j_genes)
        print('unknown_j_genes in df:', unknown_j_genes_mask.sum(),
              df[unknown_j_genes_mask][j_column].unique())
        unknown_genes_mask |= unknown_j_genes_mask


    return df[good_cdr3s_mask & (~bad_v_genes_mask) & (~unknown_genes_mask)].copy()

class BackgroundModel():
    def __init__(
        self,
        repertoire,
        factor:int=10,
        # chain:str="B",
        organism:str='human',
        v_column:str='v_call',
        j_column:str='j_call',
        cdr3nt_column:str='junction',
        cdr3aa_column:str='junction_aa', 
        verbose=False
        ):
        '''
        Class for creating synthetic backgrounds that match properties of the
        input repertoire, including V/J gene distribution, and CDR3 length.
        '''
        self.repertoire = repertoire
        self.factor = factor
        # self.chain = format_chain(chain)
        self.organism = organism
        self.junctions = None
        self.size = len(self.repertoire)
        self.n = self.size * self.factor
        self._setup_gene_ref()

        # Prepare gene columns
        self.v_column = v_column
        self.j_column = j_column
        self.cdr3nt_column = cdr3nt_column
        self.cdr3aa_column = cdr3aa_column
        # self._add_v_gene_column()
        # self._add_j_gene_column()
        # self.v_genes = self.repertoire.v_gene.unique()

        self.verbose = verbose

    def _add_v_gene_column(self):
        '''
        Add a V gene column to the repertoire file by parsing the V allele (v_call) column.
        '''
        self.repertoire['v_gene'] = self.repertoire.v_call.apply(lambda x: x.split('*')[0])

    def _add_j_gene_column(self):
        '''
        Add a J gene column to the repertoire file by parsing the J allele (j_call) column.
        '''
        self.repertoire['j_gene'] = self.repertoire.j_call.apply(lambda x: x.split('*')[0])

    def _setup_gene_ref(self):
        '''
        Setup reference database 
        !NOTE! the following V alleles have been removed from the reference:
            - TRBV7-2*03 (71)
            - TRBV7-3*04 (76)
            - TRBV15*03 (22)
            - TRBV4-3*02 (48)
        '''
        func_v_genes = pd.read_csv(join(MODELS,'functional_V_genes.csv'), index_col=[0])
        func_j_genes = pd.read_csv(join(MODELS,'functional_J_genes.csv'), index_col=[0])
        self.v_gene_ref = {v:func_v_genes[func_v_genes.gene==v].allele.to_list() for v in func_v_genes.gene.unique()}
        self.j_gene_ref = {j:func_j_genes[func_j_genes.gene==j].allele.to_list() for j in func_j_genes.gene.unique()}
        self.v_allele_ref = dict(zip(func_v_genes.allele, func_v_genes.index))
        self.j_allele_ref = dict(zip(func_j_genes.allele, func_j_genes.index))


    def _parse_junctions(self, chain):
        '''
        setup for the "resample_background_tcrs" function by parsing
        the V(D)J junctions in the foreground tcr set

        returns a dataframe with info
        '''

        chain = format_chain(chain)

        # tcrdist parsing function expects paired tcrs as list of tuples of tuples
        tcr_columns = [
            self.v_column, 
            self.j_column, 
            self.cdr3aa_column, 
            self.cdr3nt_column
            ]
        if 'junction' in self.repertoire.columns:
            self.repertoire[self.cdr3nt_column] = self.repertoire[self.cdr3nt_column].str.lower()
        else:
            self.repertoire['cdr3a_nucseq'] = self.repertoire.cdr3a_nucseq.str.lower()
            self.repertoire['cdr3b_nucseq'] = self.repertoire.cdr3b_nucseq.str.lower()
        if chain in ['A','G']:
            print(chain)
            if self.repertoire.v_call.apply(lambda x: x[:3]).nunique() > 1:
                ag = self.repertoire[self.repertoire['locus'] == f'TR{chain}']
                ag = ag[tcr_columns]
                tcr_tuples = ag.itertuples(name=None, index=None)
            else:
                print('Single chain detected.')
                tcr_tuples = self.repertoire[tcr_columns].itertuples(name=None, index=None)
            tcr_tuples = zip(tcr_tuples, itertools.repeat(None))
        elif chain in ['B','D']:
            print(chain)
            if self.repertoire.v_call.apply(lambda x: x[:3]).nunique() > 1:
                bd = self.repertoire[self.repertoire['locus'] == f'TR{chain}']
                bd = bd[tcr_columns]
                tcr_tuples = bd.itertuples(name=None, index=None)
            else:
                print('Single chain detected.')
                tcr_tuples = self.repertoire[tcr_columns].itertuples(name=None, index=None)
            tcr_tuples = zip(itertools.repeat(None), tcr_tuples)
        elif chain in ['AB','GD']:
            print(chain)
            # Check formatting
            airr_cols = ['v_call','j_call','junction_aa','junction']
            if all([i in self.repertoire.columns for i in airr_cols]):
                print('Detected AIRR format')
                ag = self.repertoire[self.repertoire['locus'] == f'TR{list(chain)[0]}']
                ag = ag[tcr_columns]
                bd = self.repertoire[self.repertoire['locus'] == f'TR{list(chain)[1]}']
                bd = bd[tcr_columns]
            else:
                print('No AIRR format detected, proceeding with paired TCRdist format.')
                ag = self.repertoire[['va','ja','cdr3a','cdr3a_nucseq']]
                bd = self.repertoire[['vb','jb','cdr3b','cdr3b_nucseq']]
            tcr_tuples_ag = ag.itertuples(name=None, index=None)
            tcr_tuples_bd = bd.itertuples(name=None, index=None)
            tcr_tuples = zip(tcr_tuples_ag, tcr_tuples_bd)

        self.junctions = parse_tcr_junctions(self.organism, list(tcr_tuples))
        #junctions = add_vdj_splits_info_to_junctions(junctions)
        return self.junctions

    def resample_background_tcrs(
        self,
        chain,
        preserve_vj_pairings=False, # consider setting True for alpha chain
        return_nucseq_srcs=False, # for debugging
        verbose=False
        ):
        '''
        Resample shuffled background sequences, number will be equal to size of
        foreground repertore, ie junctions.shape[0]

        junctions is a dataframe with information about the V(D)J junctions in the
        foreground tcrs. Created by the function "_parse_junctions()".

        returns a list of tuples [(v,j,cdr3aa,cdr3nt), ...] of length = junctions.shape[0]
        '''

        chain = format_chain(chain)

        if self.junctions is None:
            self._parse_junctions()

        multiplier = 3 # so we have enough to match distributions
        bg_tcr_tuples, src_junction_indices = resample_shuffled_tcr_chains(
            self.organism, multiplier * self.junctions.shape[0], chain, self.junctions,
            preserve_vj_pairings=preserve_vj_pairings,
            return_src_junction_indices=True
            )

        fg_nucseq_srcs = list(self.junctions[f'cdr3{chain.lower()}_nucseq_src'])

        resamples = []
        for tcr, inds in zip(bg_tcr_tuples, src_junction_indices):
            if len(resamples)%500000==0:
                if self.verbose:
                    print(f'{Background.resample_background_tcrs.__name__}: build nucseq_srclist', len(resamples),
                        len(bg_tcr_tuples))
            v,j,cdr3aa,cdr3nt = tcr
            v_nucseq = fg_nucseq_srcs[inds[0]]
            j_nucseq = fg_nucseq_srcs[inds[1]]
            nucseq_src = v_nucseq[:inds[2]] + j_nucseq[inds[2]:]
            assert len(tcr[3]) == len(nucseq_src)   
            resamples.append((v, j, cdr3aa, cdr3nt, nucseq_src))


        # try to match lengths first
        fg_lencounts = Counter(len(x) for x in self.junctions['cdr3'+chain.lower()])

        N = self.junctions.shape[0]
        good_resamples = resamples[:N]
        bad_resamples = resamples[N:]

        bg_lencounts = Counter(len(x[2]) for x in good_resamples)
        all_lencounts = Counter(len(x[2]) for x in resamples)
        if not all(all_lencounts[x]>=fg_lencounts[x] for x in fg_lencounts):
            if self.verbose:
                print('dont have enough of all lens')

        tries = 0
        too_many_tries = 10*self.junctions.shape[0]

        while True:
            tries += 1
            # pick a good tcr with a bad length and a bad tcr with a good length, swap them
            ii = np.random.randint(0,len(bad_resamples))
            iilen = len(bad_resamples[ii][2])
            if bg_lencounts[iilen] < fg_lencounts[iilen]: # too few of len=iilen
                tries = 0
                while True:
                    tries += 1
                    jj = np.random.randint(0,len(good_resamples))
                    jjlen = len(good_resamples[jj][2])
                    if bg_lencounts[jjlen] > fg_lencounts[jjlen] or tries>too_many_tries:
                        break

                if tries>too_many_tries:
                    print('WARNING too_many_tries1:', tries)
                    break
                # swap!
                dev = sum(abs(fg_lencounts[x]-bg_lencounts[x]) for x in fg_lencounts)
                #print(f'swap: {dev} {iilen} {jjlen} {tries} {too_many_tries}')
                tmp = good_resamples[jj]
                good_resamples[jj] = bad_resamples[ii]
                bad_resamples[ii] = tmp
                bg_lencounts[iilen] += 1
                bg_lencounts[jjlen] -= 1

                # are we done? if so, break out
                if all((fg_lencounts[x]<=bg_lencounts[x] or
                        bg_lencounts[x]==all_lencounts[x]) for x in fg_lencounts):
                    break
                else:
                    pass
                    # print('devs:', end=' ')
                    # for ii in range(100):
                    #     if fg_lencounts[ii] != bg_lencounts[ii]:
                    #         print(ii, fg_lencounts[ii]-bg_lencounts[ii], end=' ')
                    # print()

        assert len(good_resamples) == N

        fg_ncounts = Counter(x.count('N') for x in fg_nucseq_srcs)
        bg_ncounts = Counter(x[4].count('N') for x in good_resamples)

        if chain == 'B':
            desirable_Ncounts = [0,1]
        else:
            desirable_Ncounts = [x for x in range(10)
                                if bg_ncounts[x] < 0.9 * fg_ncounts[x]]
        if self.verbose:
            print('desirable_Ncounts:', desirable_Ncounts)
        bad_resamples = [x for x in bad_resamples if x[4].count('N') in desirable_Ncounts]

        # now try to match the N insertion distributions, while preserving the
        # length distributions
        bad_ncounts = Counter(x[4].count('N') for x in bad_resamples)
        all_ncounts = Counter(x[4].count('N') for x in resamples)

        tries = 0
        too_many_tries = 10*self.junctions.shape[0]
        too_many_inner_tries = 1000

        for num in desirable_Ncounts:
            if self.verbose:
                print('Ns:', num, 'fg_ncounts:', fg_ncounts[num],
                    'bg_ncounts:', bg_ncounts[num], 'bad_ncounts:', bad_ncounts[num],
                    'sum:', bg_ncounts[num] + bad_ncounts[num])
            if fg_ncounts[num] > all_ncounts[num]:
                if self.verbose:
                    print(f'{SyntheticBackground.resample_background_tcrs.__name__} dont have enough Ns:',
                        num, fg_ncounts[num], '>', all_ncounts[num])

        for ii in range(len(bad_resamples)):
            tries += 1
            if tries > too_many_tries:
                print('WARNING too_many_tries2:', tries)
                break
            iilen = len(bad_resamples[ii][2])
            if bg_lencounts[iilen]==0:
                continue
            iinc = bad_resamples[ii][4].count('N')
            assert iinc in desirable_Ncounts # now by construction
            if iinc in desirable_Ncounts and fg_ncounts[iinc] > bg_ncounts[iinc]:
                # find good_resamples with same len, elevated nc
                inner_tries = 0
                while True:
                    inner_tries += 1
                    jj = np.random.randint(0,len(good_resamples))
                    jjlen = len(good_resamples[jj][2])
                    if jjlen != iilen:
                        continue
                    jjnc = good_resamples[jj][4].count('N')
                    if bg_ncounts[jjnc] > fg_ncounts[jjnc]:
                        break
                    if inner_tries > too_many_inner_tries:
                        break
                if inner_tries > too_many_inner_tries:
                    tries += inner_tries//10
                    continue

                #print('swap:', iinc, jjnc, iilen, fg_ncounts[iinc]-bg_ncounts[iinc],
                #      tries)
                tmp = good_resamples[jj]
                good_resamples[jj] = bad_resamples[ii]
                bad_resamples[ii] = tmp
                bg_ncounts[iinc] += 1
                bg_ncounts[jjnc] -= 1
                bad_ncounts[iinc] -= 1
                bad_ncounts[jjnc] += 1
                if all(bad_ncounts[x] == 0 for x in desirable_Ncounts):
                    if self.verbose:
                        print('ran out of desirable_Ncounts:', desirable_Ncounts)
                    break

        if self.verbose:
            print('final Ndevs:', end=' ')
        for ii in range(100):
            if fg_ncounts[ii] != bg_ncounts[ii]:
                if self.verbose:
                    print(ii, fg_ncounts[ii]-bg_ncounts[ii], end=' ')
        # print()
        if self.verbose:
            print('final Ldevs:', end=' ')
        for ii in range(100):
            if fg_lencounts[ii] != bg_lencounts[ii]:
                if self.verbose:
                    print(ii, fg_lencounts[ii]-bg_lencounts[ii], end=' ')
        # print()

        good_tcrs = [x[:4] for x in good_resamples]
        if return_nucseq_srcs:
            good_nucseq_srcs = [x[4] for x in good_resamples]
            return good_tcrs, good_nucseq_srcs
        else:
            return good_tcrs

    def process_background(self, chain) -> pd.DataFrame:
        return pd.DataFrame(self.resample_background_tcrs(chain=chain))
    

    def shuffle(self, chain):
        """
        Shuffles T-cell receptor (TCR) sequences for a specified chain or pair of chains.

        This function generates a background dataset of TCR sequences by resampling and shuffling existing sequences.
        It supports shuffling for individual chains ('A', 'B', 'G', 'D') or a pair of chains ('AB', 'GD').
        For individual chains, it parses junctions and processes the background for the specified chain.
        For the 'AB' pair, it resamples background TCRs for both chains while preserving V-J pairings (alpha),
        and filters out sequences with bad genes or CDR3 regions.

        Parameters:
        - chain (str): The TCR locus to shuffle. Accepted values are 'A', 'B', 'G' or 'D' for single chains,
         and 'AB' or 'GD' for paired chains.

        Returns:
        - pd.DataFrame: A DataFrame containing the shuffled TCR sequences. The columns of the DataFrame
        depend on the input chain(s). For individual chains, the columns are V gene, J gene, CDR3 amino acid sequence,
        and CDR3 nucleotide sequence. For 'AB' pairs, the columns include these for both chains A and B.
        """

        if chain in ['A','B','G','D']:
            self._parse_junctions(chain=chain)
            background = [self.process_background(chain=chain) for i in range(self.factor)]
            background = pd.concat(background)
            background.columns = [self.v_column, self.j_column, self.cdr3aa_column, self.cdr3nt_column]
            return background

        chain = format_chain(chain)
        if chain == 'AB':
            self._parse_junctions(chain=chain)
            dfl = []
            for r in range(self.factor):
                # Get alpha and beta background
                tups_a = self.resample_background_tcrs(chain='A',preserve_vj_pairings=True)
                tups_b = self.resample_background_tcrs(chain='B')
                bg_tcrs = pd.DataFrame([
                    dict(va=va, ja=ja, cdr3a=cdr3a, cdr3a_nucseq=cdr3a_nucseq,
                        vb=vb, jb=jb, cdr3b=cdr3b, cdr3b_nucseq=cdr3b_nucseq)
                        for ((va,ja,cdr3a,cdr3a_nucseq), (vb,jb,cdr3b,cdr3b_nucseq)) in zip(tups_a, tups_b)
                        ])
                # Remove any TCRs with bad (ORF, pseudo) genes or non-AA characters in CDR3
                bg_tcrs = filter_out_bad_genes_and_cdr3s(
                    bg_tcrs, 'va', 'cdr3a', 'human', 'A', j_column='ja')
                bg_tcrs = filter_out_bad_genes_and_cdr3s(
                    bg_tcrs, 'vb', 'cdr3b', 'human', 'B', j_column='jb')
                dfl.append(bg_tcrs)
            return pd.concat(dfl).reset_index(drop=True)

    def olga(self, chain):
        '''
        Generates a background repertoire using the OLGA model.
        Currently only supports alpha-beta TCRs since there is
        no OLGA model for gamma-delta TCRs.
        '''
        # Standardize chain selection format
        chain = format_chain(chain)
        # Single chain
        if chain in ['A','B']:
            nsample = self.repertoire[self.repertoire['locus']==f'TR{chain}'].shape[0] * self.factor
            return generate_sequences(nsample, chain=chain)
        # Paired chains
        elif chain == 'AB':
            nsample = self.repertoire[self.repertoire['locus']==f'TRA'].shape[0] * self.factor
            alpha = generate_sequences(nsample, chain='A')
            alpha = alpha[['junction','junction_aa','v_call','j_call']]
            alpha.columns = ['cdr3a_nucseq','cdr3a','va','ja']
            beta = generate_sequences(nsample, chain='B')
            beta = beta[['junction','junction_aa','v_call','j_call']]
            beta.columns = ['cdr3b_nucseq','cdr3b','vb','jb']
            paired = pd.concat([alpha,beta],axis=1)
            paired = paired[['va', 'ja', 'cdr3a', 'cdr3a_nucseq', 'vb', 'jb', 'cdr3b','cdr3b_nucseq']]
            return paired


    # def test_background(self):

    #     if self.chain in ['A','B','G','D']:
    #         return [self.process_background() for i in range(self.factor)]
    #     else:
    #         if chain == 'AB':
    #             a = 'A'
    #             b = 'B'
    #         print(a,b)
    #         dfl = []
    #         for r in range(self.factor):
    #             tups_a = self.resample_background_tcrs(chain='A')
    #             tups_b = self.resample_background_tcrs(chain='B')
    #             bg_tcrs = pd.DataFrame([
    #                 dict(va=va, ja=ja, cdr3a=cdr3a, cdr3a_nucseq=cdr3a_nucseq,
    #                     vb=vb, jb=jb, cdr3b=cdr3b, cdr3b_nucseq=cdr3b_nucseq)
    #                     for ((va,ja,cdr3a,cdr3a_nucseq), (vb,jb,cdr3b,cdr3b_nucseq)) in zip(tups_a, tups_b)
    #                     ])
    #             bg_tcrs = filter_out_bad_genes_and_cdr3s(
    #                 bg_tcrs, 'va', 'cdr3a', 'human', 'A', j_column='ja')
    #             bg_tcrs = filter_out_bad_genes_and_cdr3s(
    #                 bg_tcrs, 'vb', 'cdr3b', 'human', 'B', j_column='jb')
    #             dfl.append(bg_tcrs)
    #         return pd.concat(dfl).reset_index(drop=True)

    # def shuffled_background(self, num_workers=1, destination=None):
    #     '''
    #     Creates a background repertoire by reshuffling the input repertoire
    #     n times (n = factor).

    #     num_workers: number of CPUs allocated
    #     '''
    #     # If num_workers > 1, use multiprocessing
    #     if num_workers > 1:
    #         import multiprocessing as mp 
    #         with mp.Pool(processes=num_workers) as pool:
    #             backgrounds = pool.starmap(self.process_background, [() for _ in range(self.factor)])
    #     else:
    #         backgrounds = [self.process_background() for i in range(self.factor)]
    #     # Combine reshuffled repertoires
    #     bg = pd.concat(backgrounds)
    #     bg.columns = [self.v_column, self.j_column, self.cdr3aa_column, self.cdr3nt_column]
    #     # Option: save to disk
    #     if destination is not None:
    #         bg.to_csv(destination, sep="\t", index=False)
    #     return bg

    # def batch_shuffling(self, destination):
    #     '''
    #     Performs background shuffling in batches and iteratively stores
    #     output on disk. 
    #     '''
    #     for i in range(self.factor):
    #         background = self.process_background()
    #         if i == 0:
    #             background.columns = [self.v_column, self.j_column, self.cdr3aa_column, self.cdr3nt_column]
    #             background.to_csv(destination, sep="\t", index=False)
    #         else:
    #             background.to_csv(destination, sep="\t", mode="a", index=False, header=False)


import pandas as pd
import numpy as np

def get_metrics(df, target_col="responding"):
    fp = 0
    tp = 0
    fpr = []
    tpr = []
    for i in df[target_col]:
        if i == False:
            fp += 1
        else:
            tp += 1
        fpr.append(fp)
        tpr.append(tp)
    N = df[target_col].value_counts()[False]
    P = df[target_col].value_counts()[True]
    fpr = [i/N for i in fpr]
    tpr = [i/P for i in tpr]
    return fpr, tpr

def annotate_results(df, responding, target_col):
    if "v_call" in responding.columns:
        df = df.merge(responding, on=["junction_aa","v_call"], how="left")
    else:
        if "v_gene" not in df.columns:
            df["v_gene"] = df.v_call.apply(lambda x: x.split("*")[0])
            df = df.merge(responding, on=["junction_aa","v_gene"], how="left")
    df[target_col] = df[target_col].fillna(False)
    # df["responding"] = df.responding.map({"red":True,"grey":False})
    return df

def identify_neighbors(n, d=1):
    if n > d:
        return False
    else:
        return True

def get_distance_from_responding(df):
    lds = []
    for i in range(len(df)):
        seq = df.iloc[i]
        cdr3 = seq["junction_aa"]
        v = seq["v_gene"]
        vsub = responding[responding.v_gene==v]
        try:
            lds.append(min([levenshtein_distance(cdr3,j) for j in vsub.junction_aa]))
        except ValueError:
            lds.append(None)
    df['ld'] = lds
    df["hit"] = df.ld.apply(lambda x: identify_neighbors(x))
    return df


import pandas as pd
import random
from os.path import dirname, abspath, join

DIR = dirname(abspath(__file__))
vdjdb_location = join(DIR,'data/vdjdb/vdjdb.txt')

# def sample_cdr3s(s:int, background:str=FILENAME, aa_col:str='junction_aa') -> list:
#     '''
#     Randomly sample s number of cdr3s from the background distribution.
#     '''
#     n = sum(1 for line in open(FILENAME)) - 1 # number of records in file (excludes header)
#     skip = sorted(random.sample(range(1,n+1),n-s)) #the 0-indexed header will not be included in the skip list
#     return pd.read_csv(background, sep='\t', skiprows=skip)[aa_col].to_list()

# def sample_tcrs(s:int, background:str=FILENAME, cols:list=['junction_aa','v_call','j_call']):
#     n = sum(1 for line in open(FILENAME)) - 1 # number of records in file (excludes header)
#     skip = sorted(random.sample(range(1,n+1),n-s)) #the 0-indexed header will not be included in the skip list
#     return pd.read_csv(background, sep='\t', skiprows=skip)[cols]

# def load_yfv_responding():
#     responding = pd.read_csv("./analysis/background_testing/yfv/yfv_responding.txt", sep="\t")
#     responding = responding[["bestVGene","CDR3.amino.acid.sequence","donor"]].drop_duplicates()
#     responding.columns = ["v_gene","junction_aa","donor"]
#     responding["responding"] = True
#     return responding

def load_vdjdb(chain="B", organism="human", exclude_10x=True, concise=True):
    
    vdjdb = pd.read_csv(vdjdb_location, sep="\t")
    vdjdb.columns = [i.replace(".","_") for i in vdjdb.columns]

    ref_map = {
        "cdr3":"junction_aa",
        "v_segm":"v_call",
        "j_segm":"j_call",
    }

    vdjdb = vdjdb.rename(columns=ref_map)
    
    if chain.upper() == "B":
        vdjdb = vdjdb[vdjdb.gene=="TRB"]
    else:
        vdjdb = vdjdb[vdjdb.gene=="TRA"]
    
    if organism.lower() == "human":
        vdjdb = vdjdb[vdjdb.species=="HomoSapiens"]
    elif organism.lower() == "mouse":
        vdjdb = vdjdb[vdjdb.species=="MusMusculus"]
    else:
        pass

    if exclude_10x:
        tenx_id = vdjdb["reference_id"].value_counts().index[0]
        vdjdb = vdjdb[vdjdb["reference_id"]!=tenx_id]

    if concise:
        vdjdb = vdjdb[["junction_aa","v_call","j_call","antigen_epitope","antigen_species"]]
    
    vdjdb = vdjdb.drop_duplicates()
    
    return vdjdb.reset_index(drop=True)



## LAZY-- some basic imports and shared params
##
##
from glob import glob
from os import popen, system, chdir, remove, getcwd, mkdir
from os.path import exists, isdir, isfile
import math
from math import floor,sqrt
from sys import stderr,argv,exit
import random

#from ..convert_svg_to_png import convert_svg_to_png, MONOSPACE_FONT_FAMILY
MONOSPACE_FONT_FAMILY = 'DejaVu Sans Mono'
def convert_svg_to_png(*args,**kwargs):
    print('convert_svg_to_png is undefined')
    exit()

# combo dbs for tr and ig
db_file = 'combo_xcr.tsv'
#db_file = 'combo_db.tsv'
#db_file = 'alphabeta_db.tsv'

# should we use the crazy complicated tcr-dist scheme for picking a 'representative' gene for counting?
CLASSIC_COUNTREPS = False

## naming scheme for the gene segment types, occasionally useful for iterating
segtypes_uppercase = ['VA','JA','VB','JB']
segtypes_lowercase = ['va','ja','vb','jb']






# def Log(s): ## silly legacy helper function
#     stderr.write(s)
#     if s and not s.endswith('\n'):
#         stderr.write('\n')





## trees from distances -- this was all written ages ago before scipy/sklearn

import string
from os import popen,system,getcwd
import sys
from math import floor,log,exp
#from popen2 import popen2
#from os.path import exists
#from glob import glob

## for making a small glyph that shows the protein-dna interactions
## in a cluster


def IsALeaf(node): return (node[0] == node[1])


def Average_score (leaf_list, leaf_scores, percentile):
    ## negative percentile is signal to use straight average...
    ls = []
    for leaf in leaf_list: ## some of leaf_scores[leaf] could be empty lists...
        ls = ls + leaf_scores[leaf]
    if not ls: ## no score information for this set of leaves
        return None
    elif percentile >= 0:
        assert type(percentile) is int
        ls.sort()
        pos = (percentile * len(ls) ) / 100
        if pos==len(ls): pos = len(ls)-1

        return ls[pos]
    else:
        return sum( ls ) / float( len( ls ) )

class CallAverageScore:
    def __init__( self, percentile ):
        self.percentile = percentile
    def __call__( self, leaf_list, leaf_scores ):
        return Average_score( leaf_list, leaf_scores, self.percentile )



def Make_tree_new(
        distance,
        num_leaves,
        Update_distance_matrix,
        leaf_scores,
        Compute_average_score
):
    # Compute_average_score has to be callable with two args: leaf_list and leaf_scores
    N = num_leaves

    nodes = []
    for i in range(N):
        nodes.append( (i,i,0.0,Compute_average_score([i],leaf_scores) ) )

    for i in range(N): ## initialize distance matrix
        for j in range(N):
            distance[(nodes[i],nodes[j])] = distance[(i,j)]


    while N>1:

        ## find two closest nodes and join them
        min_d = 100000

        for ii, i in enumerate(nodes):
            for jj, j in enumerate(nodes):
                #if i<=j:continue # used to work
                if ii<=jj:continue
                if distance[(i,j)] < min_d:
                    min_d = distance[(i,j)]
                    n1 = i
                    n2 = j

        ##         print "num_nodes: %d   Joining: %s and %s   Distance: %7.3f\n"\
        ##               %(N,Show_small(n1),Show_small(n2),min_d)

        new_node_score = Compute_average_score( Node_members(n1)+Node_members(n2),leaf_scores)

        new_node = (n1,n2,min_d,new_node_score)


        ## update the distances
        Update_distance_matrix (new_node,nodes,distance)

        ## update the node_list
        nodes.append(new_node)
        del nodes[ nodes.index(n1)]
        del nodes[ nodes.index(n2)]

        N = N-1

    return nodes[0]

# def Make_tree_zero_scores( distance, num_leaves, Update_distance_matrix ):
#     N = num_leaves

#     nodes = []
#     for i in range(N):
#         nodes.append( (i,i,0.0,0.0) )

#     for i in range(N): ## initialize distance matrix
#         for j in range(N):
#             distance[(nodes[i],nodes[j])] = distance[(i,j)]


#     while N>1:

#         ## find two closest nodes and join them
#         min_d = 100000

#         for i in nodes:
#             for j in nodes:
#                 if i<=j:continue
#                 if distance[(i,j)] < min_d:
#                     min_d = distance[(i,j)]
#                     n1 = i
#                     n2 = j

#         new_node = (n1,n2,min_d,0.0)

#         ## update the distances
#         Update_distance_matrix (new_node,nodes,distance)

#         ## update the node_list
#         nodes.append(new_node)
#         del nodes[ nodes.index(n1)]
#         del nodes[ nodes.index(n2)]

#         N = N-1
#     return nodes[0]

## the old way
def Make_tree(distance,num_leaves,Update_distance_matrix,leaf_scores,percentile):
    func = CallAverageScore( percentile )
    return Make_tree_new( distance, num_leaves, Update_distance_matrix, leaf_scores, func )

def Copy_tree_update_scores( old_tree, leaf_scores, Compute_average_score ):
    members = Node_members( old_tree )
    score = Compute_average_score( members, leaf_scores )
    if IsALeaf( old_tree ):
        return ( old_tree[0], old_tree[1], old_tree[2], score )
    else:
        return ( Copy_tree_update_scores( old_tree[0], leaf_scores, Compute_average_score ),
                 Copy_tree_update_scores( old_tree[1], leaf_scores, Compute_average_score ),
                 old_tree[2], score )


def Show_tree(tree,names):
    if IsALeaf(tree):
        return names [tree[0]]
    else:
        return '('+Show_tree(tree[0],names)+':'+str(float(tree[2])/2)+','+\
               Show_tree(tree[1],names)+':'+str(float(tree[2])/2)+')'

def Show_small(tree):
    if IsALeaf(tree):
        return repr(tree[0])
    else:
        return '('+Show_small(tree[0])+','+Show_small(tree[1])+')'

def Node_members(node):
    if IsALeaf(node):
        return [node[0]]
    else:
        l1 = Node_members( node[0] )
        l2 = Node_members( node[1] )
        if min(l1)<min(l2):
            return l1+l2
        else:
            return l2+l1


def Update_distance_matrix_AL(new_node,old_nodes,distances): ## average linkage
    n1 = new_node[0]
    n2 = new_node[1]

    l1 = Node_members(new_node)
    distances [ (new_node,new_node)] = 0.0

    for n in old_nodes:
        if n==n1 or n==n2:continue
        l2 = Node_members(n)

        avg = 0.0
        count = 0
        for i in l1:
            for j in l2:
                assert i!=j
                avg = avg+ distances[(i,j)]
                count = count + 1

        distances[(n,new_node)] = avg/count
        distances[(new_node,n)] = avg/count

    return

def Update_distance_matrix_AL_GEOM(new_node,old_nodes,distances):
    dl = list(distances.values())
    dl.sort()
    for i in dl:
        if i!=0:
            min_log = log(i) - 3 ## closer than the closest non-id pair
            break

    n1 = new_node[0]
    n2 = new_node[1]

    l1 = Node_members(new_node)
    distances [ (new_node,new_node)] = 0.0

    for n in old_nodes:
        if n==n1 or n==n2:continue
        l2 = Node_members(n)

        avg = 0.0
        count = 0
        for i in l1:
            for j in l2:
                d = distances[(i,j)]
                assert i!=j
                count = count + 1
                if d == 0.0:
                    avg = avg + min_log
                else:
                    avg = avg + log( d )

        distances[(n,new_node)] = exp( avg / count )
        distances[(new_node,n)] = exp( avg / count )

    return

def Update_distance_matrix_SL(new_node,old_nodes,distances): ## single linkage
    n1 = new_node[0]
    n2 = new_node[1]

    l1 = Node_members(new_node)
    distances [ (new_node,new_node)] = 0.0

    for n in old_nodes:
        if n==n1 or n==n2:continue
        l2 = Node_members(n)

        min_d = 1000
        count = 0
        for i in l1:
            for j in l2:
                assert i!=j
                min_d = min(min_d, distances[(i,j)])

        distances[(n,new_node)] = min_d
        distances[(new_node,n)] = min_d

    return

def Center(tree,node_position,sizes,use_sizes_as_weights=False):
    l = Node_members(tree)
    pos = 0.0
    total_weight=0.0
    for i in l:
        if use_sizes_as_weights:
            pos = pos + sizes[i]*node_position[i]
            total_weight +=sizes[i]
        else:
            pos = pos + node_position[i]
            total_weight+= 1
    pos = pos / total_weight
    return pos


def Size(tree,sizes):
    if IsALeaf(tree):
        return sizes[tree[0]]
    else:
        return Size(tree[0],sizes)+Size(tree[1],sizes)


def Fig_tree(tree,node_position,sizes,use_sizes_as_weights=False): ## edge = [ [x0,y0], [x1,y1], score, size]
    if IsALeaf(tree):
        return []
    else:

        rmsd = tree[2]
        center = Center(tree,node_position,sizes,use_sizes_as_weights)

        c0 = Center(tree[0],node_position,sizes,use_sizes_as_weights)
        r0 = tree[0][2]
        score0 = tree[0][3]
        size0 = Size(tree[0],sizes)
        if IsALeaf(tree[0]):
            cluster0 = tree[0][0]
        else:
            cluster0 = -1
        e0_horizontal = [ [rmsd, c0], [r0,c0], score0, size0, cluster0]
        e0_vertical   = [ [rmsd, c0], [rmsd,center], score0, 1, cluster0]

        c1 = Center(tree[1],node_position,sizes,use_sizes_as_weights)
        r1 = tree[1][2]
        score1 = tree[1][3]
        size1 = Size(tree[1],sizes)
        if IsALeaf(tree[1]):
            cluster1 = tree[1][0]
        else:
            cluster1 = -1
        e1_horizontal = [ [rmsd, c1], [r1, c1], score1, size1 , cluster1]
        e1_vertical   = [ [rmsd, c1], [rmsd,center], score1, 1, cluster1]

        return [ e0_vertical,e0_horizontal,e1_vertical,e1_horizontal] + \
               Fig_tree( tree[0],node_position,sizes,use_sizes_as_weights ) + \
               Fig_tree( tree[1],node_position,sizes,use_sizes_as_weights )

def Node_labels(tree,sizes,node_position,use_sizes_as_weights=False):
    if IsALeaf(tree):return []
    else:
        pos = [tree[2],Center(tree,node_position,sizes,use_sizes_as_weights)]
        size = 0
        for leaf in Node_members(tree):
            size = size+sizes[leaf]
        return [ [ repr(size), pos] ] + \
               Node_labels(tree[0],sizes,node_position,use_sizes_as_weights) + \
               Node_labels(tree[1],sizes,node_position,use_sizes_as_weights)



## return the y-coordinates of the different clusters

def Canvas_tree(
        tree,
        names,
        sizes,
        upper_left,
        lower_right,
        branch_width_fraction,
        plotter,
        label_singletons = False,
        label_internal_nodes = True,
        font=None,
        score_range_for_coloring=None,
        vertical_line_width = 1,
        show_colorful_rmsd_bar = False,
        rmsd_bar_label_stepsize=1,
        rmsd_bar_label_fontsize=10,
        force_min_rmsd=None,
        verbose=False,
):
    ## score_range_for_coloring is a tuple:(mn,mx)
    if score_range_for_coloring: assert len(score_range_for_coloring) == 2

    assert upper_left[0] < lower_right[0]
    assert upper_left[1] < lower_right[1]

    x0,y0 = upper_left
    x1,y1 = lower_right

    #plot_width  = x1-x0
    plot_height = y1-y0

    ## now we are assuming the origin is the top-left corner, like in svg

    ## plot_width and plot_height in pixels

    ## plotter has methods:
    ## .make_line ( [x0,y0], [x1,y1], line_width, normalized_score, extra_tag)
    ## .make_text (text,  [x,y], font)

    branch_width_pixels = plot_height * branch_width_fraction
    #branch_width_pixels = min(100,plot_height/5)

    ## allocate widths for branches; widths measure cluster sizes
    total = sum(sizes)
    w_factor = float( branch_width_pixels) / total
    total = 0
    for s in sizes:
        width = max(1,int(floor(0.5+ s*w_factor))) ## in pixels
        total = total+width
    remainder = plot_height - total
    cluster_width = float(remainder)/len(names) ## padding alotted to each cluster

    if verbose:
        print('branch_width_pixels: {:.2f} plot_height: {:.2f} cluster_padding: {:.3f} w_factor: {:.3f} num_clusters: {} total_members: {}'\
              .format( branch_width_pixels,plot_height,cluster_width,w_factor,len(sizes),sum(sizes)))

    ## position nodes vertically on tree
    nodes = Node_members(tree)
    node_position = {}
    mark = y1
    for i,node in enumerate(nodes):
        width = max(1,int(floor(0.5+ sizes[node] * w_factor)))
        node_position[ node ] = mark-width/2
        mark = mark - cluster_width - width

    edges = Fig_tree(tree,node_position,sizes,use_sizes_as_weights=True) ## each edge = [[x0,y0],[x1,y1],score,size,cluster]

    ## set fontsize: is this still right??

    if not font:
        font = 2*min(25, max (15, int(floor( 0.5 + (cluster_width+7.5)/10))))

    ## rescale the x-positions
    max_rmsd = tree[2]
    min_rmsd = tree[2]
    for e in edges:
        if e[0][0]>0: min_rmsd = min(min_rmsd,e[0][0])
        if e[1][0]>0: min_rmsd = min(min_rmsd,e[1][0])
    min_rmsd = max(0,min_rmsd-0.5)

    if force_min_rmsd is not None:
        min_rmsd = force_min_rmsd

    def Transform(rmsd,min_rmsd = min_rmsd, max_rmsd = max_rmsd):
        return int (floor ( 0.5 + x0 + (x1-x0) * (rmsd - min_rmsd) / (max_rmsd - min_rmsd)))


    ## rescale colors
    scores = []
    for e in edges:
        edge_score = e[2]
        if edge_score != None:
            scores.append(edge_score)

    min_score = min(scores)
    max_score = max(scores)
    if verbose:
        print('min_score:',min_score,'max_score:',max_score,score_range_for_coloring)
    if max_score == min_score:
        max_score = max_score + 1
    if score_range_for_coloring:
        min_score,max_score = score_range_for_coloring

    ## write the edges
    for e in edges:
        start = [ Transform (max(e[0][0],min_rmsd)), e[0][1]] ## rescale x-position
        stop = [ Transform (max(e[1][0],min_rmsd)), e[1][1]]
        edge_score = e[2]
        if edge_score == None:
            normalized_score = None
        else:
            normalized_score = max(0.0, min(1.0, float( e[2] - min_score)/(max_score-min_score) ) )
        line_width = max(1,int(floor(0.5+ e[3]*w_factor)))
        if e[4]>=0: ## it's a real cluster edge
            cluster = e[4]
            extra_tag = 'cluster%02d.%03d'%(cluster,sizes[cluster])
        else:
            extra_tag = 'dummy'

        if start[0] == stop[0]:
            line_width = vertical_line_width

        plotter.make_line(start,stop,line_width,normalized_score,extra_tag)


    ## show scale
    line_x0 = Transform(min_rmsd)
    line_x1 = Transform(max_rmsd)
    if show_colorful_rmsd_bar:
        line_steps = 512
        line_step_size = (line_x1-line_x0)/float(line_steps)
        for i in range( line_steps ):
            norm_score = (float(i)+0.5)/(line_steps)
            #print 'norm_score:',norm_score
            plotter.make_line([ line_x0 + i*line_step_size,y0+5],
                              [ line_x0 +(i+1)*line_step_size,y0+5],3,norm_score )
    else:
        line_ypad = 5
        line_ypad = 2
        plotter.make_line([ line_x0,y0+line_ypad],[ line_x1,y0+line_ypad],2,0.0,color='black' )


    for i in range(int(floor(min_rmsd+1)), 1+int(floor(tree[2])), rmsd_bar_label_stepsize):
        plotter.make_text( str(i), [Transform(i),y0], rmsd_bar_label_fontsize)


    #plotter.make_text( 'Colors: from blue (%7.2f) to red (%7.2f)'%(min_score,max_score),
    #                   [x0,y0+25],25)


    ## label leaves
    for i in range(len(names)):
        if sizes[i] == 1 and not label_singletons: continue
        extra_tag = 'cluster%02d.%03d'%(i,sizes[i])
        plotter.make_text(names[i],
                          [Transform(min_rmsd),node_position[i]],
                          font,extra_tag)

    ## label internal vertices with sizes
    if label_internal_nodes:
        for l in Node_labels (tree,sizes,node_position,use_sizes_as_weights=True):
            plotter.make_text(l[0], [Transform(l[1][0]),l[1][1]], font)

    return ( node_position, Transform, min_rmsd, w_factor ) ## nuisance






import numpy as np 
import faiss
from scipy.sparse import csr_matrix

def index_neighbors_manual(query, r, index):
    '''
    For each query TCR sequence i, this finds neighbors on the index that are 
    within  a radius r of each query TCR. 
    
    Directly export the lim, D, I range_search output for manual 
    manipulation of neighbors.

    Parameters
    ----------
    query
        TCRs or CDR3 sequences for which the number of neighbors on the index
        should be calculated.
    r
        Radius parameter that defines the edge of the TCR neighborhood.
    index
        Index on which the search should be performed. The index needs to be
        trained and should contain vectors to query against.
    '''
    X = index.hasher.fit_transform(query)
    X = X.astype(np.float32)
    lim, D, I = index.idx.range_search(X, thresh=r)
    return lim, D, I 

def range_search_to_csr_matrix(lims,
    D, 
    I,
    shape = None):
    """
    Convert output of faiss index.idx.range_search to a csr sparse matrix

    Parameters
    ----------
    lims
        output of index.idx.range_search, is range of positions in D and I corresponding 
        to the ith query sequence.
    D 
        neigbor distance vector output of index.idx.range_search       
    I 
        neigbor index vector output of index.idx.range_search
    shape 
        optional tuple (n_rows, n_cols) to force matrix size
    Returns
    -------
    csr_mat 
        scipy.sparse.csrmat of distance less than user specified max radius.
        zero distances will be encoded as -1 in the sparse csr_matrix 
    """
    dist_csr = list()
    for i in range(len(lims)-1):
        # <ix> is neighbor > r index position
        ix = I[lims[i]:lims[i+1]].tolist()
        # dx is neighbor > r distance
        dx = D[lims[i]:lims[i+1]].tolist()
        # convert distance to an integer
        dx_as_int = [round(x) for x in dx]
        # convert 0 dist to negative 1 for sparsity
        dx_as_int = np.array([-1 if x == 0 else x for x in dx_as_int], dtype = 'int8')
        dist_csr.append([(i,j,d) for j, d in zip(ix, dx_as_int)])
    # flatten list of tuple lists
    dist_csr = np.concatenate(dist_csr)
    row_indices = [row for row, col, val in dist_csr]
    col_indices = [col for row, col, val in dist_csr]
    values      = np.array([val for row, col, val in dist_csr], dtype = 'int8')

    if shape is None:
        csr_mat = csr_matrix((values, (row_indices, col_indices)) )
    else:
        n_rows, n_cols = shape
        csr_mat = csr_matrix((values, (row_indices, col_indices)), shape = (n_rows, n_cols), dtype = 'int8')
    return csr_mat



def query_function(q, index, n_rows, r):
  lims, D, I = index.range_search(q, thresh=r)
  csr_mat = range_search_to_csr_matrix(lims, D, I, (q.shape[0], n_rows)) 
  return csr_mat




import sys
import random
from collections import OrderedDict, Counter
import pandas as pd
from .basic import *
from . import translation
from .all_genes import all_genes, gap_character
from .genetic_code import genetic_code, reverse_genetic_code
from . import logo_tools

# set up legacy D gene nucleotide sequence dict, 1-indexed integer keys
all_trbd_nucseq = {}
for organism in all_genes:
    d_ids = sorted((id for id,g in all_genes[organism].items()
                    if g.region == 'D' and g.chain == 'B'))
    all_trbd_nucseq[ organism ] = dict(
        ((d_ids.index(id)+1, all_genes[organism][id].nucseq ) for id in d_ids))

########################################################################################################################
default_mismatch_score_for_cdr3_nucseq_probabilities = -4 ## blast is -3
default_mismatch_score_for_junction_analysis = -4 ## blast is -3

def count_matches( a,b,mismatch_score=-3 ):
    assert a[0].lower() == a[0]
    #assert a[0].upper() == a[0]
    ## from the beginning
    match_score    = 1
    best_score=0
    score=0
    num_matches = 0
    for i in range(min(len(a),len(b))):
        if ( a[i] == b[i] or
             logo_tools.nuc_match_lower_case.get((a[i],b[i]), False)):
            score += match_score
        else:
            score += mismatch_score
        if score >= best_score: ## gt OR EQUAL! take longer matched regions
            best_score = score
            num_matches = i+1
    return num_matches

def count_aa_matches( a,b,mismatch_score=-3 ):
    ## from the beginning
    assert a[0].upper() == a[0]
    match_score    = 1
    best_score=0
    score=0
    num_matches = 0
    for i in range(min(len(a),len(b))):
        if a[i] == b[i]:
            score += match_score
        else:
            score += mismatch_score
        if score >= best_score: ## gt OR EQUAL! take longer matched regions
            best_score = score
            num_matches = i+1
    return num_matches



def get_v_cdr3_nucseq( organism, v_gene):
    vg = all_genes[organism][v_gene]
    ab = vg.chain

    v_nucseq  = vg.nucseq
    v_nucseq_offset = vg.nucseq_offset
    v_nucseq = v_nucseq[ v_nucseq_offset: ]

    v_alseq = vg.alseq
    alseq_cpos = vg.cdr_columns[-1][0] -1
    #print organism, v_gene, old_v_alseq
    #print organism, v_gene, v_alseq
    numgaps = v_alseq[:alseq_cpos].count('.')
    v_cpos = alseq_cpos - numgaps
    v_nucseq = v_nucseq[3*v_cpos:] ## now v_nucseq starts with the 'C' codon


    # the following hacky adjustment is now incorporated in the dbfile
    # if organism == 'mouse':
    #     if v_gene == 'TRAV13D-1*01':
    #         #-----------------------------------
    #         #../../../tmp.blat:mismatch: V 6 imgt: a genome: t TRAV13D-1*01
    #         #tmp.whoah:whoah  6 act: t  98.7 exp: a   1.1 TRAV13D-1*01 TRAV13-1*01 620
    #         #tmp.whoah:whoah: expected: caaggtatcgtgt consensus: caaggtttcgtgt TRAV13D-1*01 620
    #         #tmp.3.whoah:whoah  6 act: t  97.4 exp: a   1.4 TRAV13D-1*01 TRAV13-1*01 642
    #         #tmp.3.whoah:whoah: expected: caaggtatcgtgt consensus: caaggtttcgtgt TRAV13D-1*01 642
    #         #tmp.la_mc.whoah:whoah  6 act: t  89.0 exp: a   7.0 TRAV13D-1*01 TRAV13-1*01 100
    #         #tmp.la_mc.whoah:whoah: expected: caaggtatcgtgt consensus: caaggtttcgtgt TRAV13D-1*01 100
    #         assert v_nucseq == 'tgtgctatggaac' ## CAM ## THIS WILL FAIL SINCE WE ADDED THIS TO THE DB...
    #         v_nucseq         = 'tgtgctttggaac' ## CAL


    return v_nucseq


def get_j_cdr3_nucseq( organism, j_gene):
    jg = all_genes[organism][j_gene]
    ab = jg.chain

    j_nucseq  = jg.nucseq
    j_nucseq_offset = jg.nucseq_offset

    ## goes up to (but not including) GXG
    num_genome_j_aas_in_loop = len( jg.cdrs[0].replace(gap_character,''))

    ## trim j_nucseq so that it extends up to the F/W position
    j_nucseq = j_nucseq[:3*num_genome_j_aas_in_loop + j_nucseq_offset]


    # the following hacky adjustments are now incorporated in the dbfile
    # if organism == 'mouse':
    #     if j_gene == 'TRAJ47*01':
    #         # -----------------------------------
    #         # ../../../tmp.blat:mismatch: J 2 imgt: c genome: g TRAJ47*01
    #         # ../../../tmp.blat:mismatch: J 24 imgt: g genome: t TRAJ47*01
    #         # tmp.whoah:whoah  2 act: g  81.9 exp: c   4.7 TRAJ47*01 TRAJ47*01 1412
    #         # tmp.whoah:whoah 24 act: t  82.7 exp: g  16.8 TRAJ47*01 TRAJ47*01 1412
    #         # tmp.whoah:whoah: expected: tgcactatgcaaacaagatgatctgt consensus: tggactatgcaaacaagatgatcttt TRAJ47*01 1412
    #         # tmp.3.whoah:whoah  2 act: g  81.6 exp: c   5.0 TRAJ47*01 TRAJ47*01 1362
    #         # tmp.3.whoah:whoah 24 act: t  82.7 exp: g  16.6 TRAJ47*01 TRAJ47*01 1362
    #         # tmp.3.whoah:whoah: expected: tgcactatgcaaacaagatgatctgt consensus: tggactatgcaaacaagatgatcttt TRAJ47*01 1362
    #         # tmp.la_mc.whoah:whoah  2 act: g  79.6 exp: c   5.3 TRAJ47*01 TRAJ47*01 113
    #         # tmp.la_mc.whoah:whoah 24 act: t  99.1 exp: g   0.9 TRAJ47*01 TRAJ47*01 113
    #         # tmp.la_mc.whoah:whoah: expected: tgcactatgcaaacaagatgatctgt consensus: tggactatgcaaacaagatgatcttt TRAJ47*01 113
    #         assert j_nucseq == 'tgcactatgcaaacaagatgatctgt' ## C at end
    #         j_nucseq         = 'tggactatgcaaacaagatgatcttt' ## F at end
    #     elif j_gene == 'TRAJ24*01':
    #         # -----------------------------------
    #         # ../../../tmp.blat:unaligned: J 0 TRAJ24*01
    #         # ../../../tmp.blat:unaligned: J 1 TRAJ24*01
    #         # ../../../tmp.blat:gapped: J 6 TRAJ24*01
    #         # tmp.whoah:whoah  2 act: c  60.3 exp: a  15.3 TRAJ24*01 TRAJ24*01 464
    #         # tmp.whoah:whoah  4 act: a  88.6 exp: c   2.8 TRAJ24*01 TRAJ24*01 464
    #         # tmp.whoah:whoah  5 act: c  93.3 exp: t   1.5 TRAJ24*01 TRAJ24*01 464
    #         # tmp.whoah:whoah  6 act: t  97.2 exp: g   1.1 TRAJ24*01 TRAJ24*01 464
    #         # tmp.whoah:whoah: expected: tgaactggccagtttggggaaactgcagttt consensus: gacaactgccagtttggggaaactgcagttt TRAJ24*01 464
    #         # tmp.3.whoah:whoah  2 act: c  60.8 exp: a  13.9 TRAJ24*01 TRAJ24*01 475
    #         # tmp.3.whoah:whoah  4 act: a  86.3 exp: c   4.2 TRAJ24*01 TRAJ24*01 475
    #         # tmp.3.whoah:whoah  5 act: c  94.5 exp: t   1.1 TRAJ24*01 TRAJ24*01 475
    #         # tmp.3.whoah:whoah  6 act: t  98.1 exp: g   1.1 TRAJ24*01 TRAJ24*01 475
    #         # tmp.3.whoah:whoah: expected: tgaactggccagtttggggaaactgcagttt consensus: gacaactgccagtttggggaaactgcagttt TRAJ24*01 475
    #         # tmp.la_mc.whoah:whoah  2 act: c  75.3 exp: a   4.3 TRAJ24*01 TRAJ24*01 93
    #         # tmp.la_mc.whoah:whoah  4 act: a  89.2 exp: c   2.2 TRAJ24*01 TRAJ24*01 93
    #         # tmp.la_mc.whoah:whoah  5 act: c  97.8 exp: t   1.1 TRAJ24*01 TRAJ24*01 93
    #         # tmp.la_mc.whoah:whoah  6 act: t  98.9 exp: g   0.0 TRAJ24*01 TRAJ24*01 93
    #         # tmp.la_mc.whoah:whoah: expected: tgaactggccagtttggggaaactgcagttt consensus: gacaactgccagtttggggaaactgcagttt TRAJ24*01 93
    #         assert j_nucseq == 'tgaactggccagtttggggaaactgcagttt'
    #         j_nucseq         = 'gacaactgccagtttggggaaactgcagttt'
    #         ## take the consensus
    #         ## given that there's an indel (and the alignment to the genome starts at j sequence position 3)
    #         ## it's hard to tell what to do at the beginning...



    return j_nucseq


def find_alternate_alleles(
        organism,
        v_gene,
        j_gene,
        cdr3_nucseq,
        min_improvement = 2
        ):
    ''' Since 10x generally doesn't provide allele (ie *01) information,
    we may want to see if there's an alternate allele that provides a
    better explanation of the cdr3 nucleotide sequence

    return new_v_gene, new_j_gene

    only take an alternate allele if its coverage of cdr3_nucseq is greater by at least
    min_improvement over the current allele
    '''
    mismatch_score = default_mismatch_score_for_junction_analysis # -4
    # want to eventually make this the default:
    #mismatch_score = -1000 # 2022-01-21  this makes more sense to me (PB)

    v_prefix = v_gene[:v_gene.index('*')+1]
    alternate_v_alleles = [ x for x in all_genes[organism] if x.startswith(v_prefix) and x != v_gene]

    j_prefix = j_gene[:j_gene.index('*')+1]
    alternate_j_alleles = [ x for x in all_genes[organism] if x.startswith(j_prefix) and x != j_gene]

    for vg in [v_gene]+alternate_v_alleles:
        v_nucseq = get_v_cdr3_nucseq( organism, vg )
        if not v_nucseq:
            #print(vg, v_nucseq)
            continue
        num_matched = count_matches(v_nucseq, cdr3_nucseq, mismatch_score)
        if vg == v_gene:
            best_v_gene = vg
            best_num_matched = num_matched
        else:
            if ( num_matched > best_num_matched and ( best_v_gene != v_gene or
                                                      num_matched - best_num_matched >= min_improvement)):
                #print('new best v_gene:', vg, best_v_gene, num_matched, best_num_matched)
                best_v_gene = vg
                best_num_matched = num_matched

    for jg in [j_gene]+alternate_j_alleles:
        j_nucseq = get_j_cdr3_nucseq( organism, jg )
        num_matched = count_matches(''.join( reversed( list( j_nucseq ) )),
                                    ''.join( reversed( list( cdr3_nucseq ))),
                                    mismatch_score)
        if jg == j_gene:
            best_j_gene = jg
            best_num_matched = num_matched
        else:
            if ( num_matched > best_num_matched and ( best_j_gene != j_gene or
                                                      num_matched - best_num_matched >= min_improvement)):
                #print('new best j_gene:', jg, best_j_gene, num_matched, best_num_matched)
                best_j_gene = jg
                best_num_matched = num_matched
    return best_v_gene, best_j_gene


def find_alternate_alleles_for_tcrs(
        organism,
        tcrs,
        min_better_ratio=10,
        min_better_count=5,
        min_improvement_to_be_called_better=2,
        verbose=False,
        very_verbose=False,
        ):
    ''' Try looking for cases where an alternate allele explains more of the cdr3_nucseq
    than the current allele (which is probably *01 since 10x usually doesn't provide alleles)

    For ones that occur sufficiently often, swap them into the tcrs where they match better
    '''
    if verbose:
        print('find_alternate_alleles_for_tcrs: num_tcrs:', len(tcrs))
    all_counts = {}
    all_new_genes = []
    for atcr, btcr in tcrs:
        new_genes = []
        for tcr in [atcr, btcr]:
            v, j, _, cdr3_nucseq = tcr
            new_v, new_j = find_alternate_alleles(
                organism, v, j, cdr3_nucseq, min_improvement=min_improvement_to_be_called_better)
            all_counts.setdefault(v, Counter())[new_v] += 1
            all_counts.setdefault(j, Counter())[new_j] += 1
            new_genes.append((new_v, new_j))
        all_new_genes.append(new_genes)

    allowed_swaps = set()
    for g, counts in all_counts.items():
        l = counts.most_common()
        if l[0][0] != g or (len(l)>1 and l[1][1] >= l[0][1]//min_better_ratio and l[1][1]>=min_better_count):
            if l[0][0] != g:
                alt_g = l[0][0]
            else:
                alt_g = l[1][0]
            if verbose:
                print('find_alternate_alleles_for_tcrs: allow alternate gene:', g, alt_g)
            allowed_swaps.add(alt_g)

    new_tcrs = []
    for paired_tcr, paired_new_genes in zip(tcrs, all_new_genes):
        new_tcr = []
        for tcr, new_genes in zip(paired_tcr, paired_new_genes):
            v, j, cdr3, cdr3_nucseq = tcr
            new_v, new_j = new_genes
            if new_v in allowed_swaps:
                v = new_v
            if new_j in allowed_swaps:
                j = new_j
            new_tcr.append((v, j, cdr3, cdr3_nucseq))
        new_tcr = tuple(new_tcr)
        if new_tcr != paired_tcr:
            if very_verbose:
                print('find_alternate_alleles_for_tcrs: old_tcr:', paired_tcr)
                print('find_alternate_alleles_for_tcrs: new_tcr:', new_tcr)
        new_tcrs.append(new_tcr)
    return new_tcrs




def analyze_junction(
        organism,
        v_gene,
        j_gene,
        cdr3_protseq,
        cdr3_nucseq,
        force_d_id=0,
        return_cdr3_nucseq_src=False,
        mismatch_score=default_mismatch_score_for_junction_analysis,
):
    assert cdr3_nucseq == cdr3_nucseq.lower() # confirm lowercase

    #print organism, v_gene, j_gene, cdr3_protseq, cdr3_nucseq
    #assert v_gene.startswith('TR') #and v_gene[2] == j_gene[2]

    # if it's not all standard bases, we need to change how we do the
    # D gene finding...
    cdr3_nucseq_has_other = any(x not in 'acgt' for x in cdr3_nucseq)

    ab = all_genes[organism][v_gene].chain
    v_nucseq = get_v_cdr3_nucseq( organism, v_gene )
    j_nucseq = get_j_cdr3_nucseq( organism, j_gene )
    ## how far out do we match
    num_matched_v = count_matches( v_nucseq, cdr3_nucseq, mismatch_score)

    num_matched_j = count_matches( ''.join( reversed( list( j_nucseq ) )),
                                   ''.join( reversed( list( cdr3_nucseq ))),
                                   mismatch_score)


    if num_matched_v + num_matched_j > len(cdr3_nucseq):
        ## some overlap!
        extra = num_matched_v + num_matched_j - len(cdr3_nucseq )
        fake_v_trim = extra//2 ## now deterministic
        fake_j_trim = extra - fake_v_trim
        num_matched_v -= fake_v_trim
        num_matched_j -= fake_j_trim

    assert num_matched_v + num_matched_j <= len(cdr3_nucseq)

    if num_matched_v + num_matched_j == len(cdr3_nucseq):
        nseq = ''
    else:
        nseq = cdr3_nucseq[num_matched_v:len(cdr3_nucseq)-num_matched_j]

    ncount = [1] * len(cdr3_nucseq)
    cdr3_nucseq_src = ['N'] * len(cdr3_nucseq)
    for i in range(num_matched_v):
        ncount[i] = 0
        cdr3_nucseq_src[i] = 'V'
    for i in range(num_matched_j):
        ncount[-1-i] = 0
        cdr3_nucseq_src[-1-i] = 'J'

    assert num_matched_v + num_matched_j + len(nseq) == len(cdr3_nucseq)

    v_trim = len(v_nucseq)-num_matched_v
    j_trim = len(j_nucseq)-num_matched_j

    assert (len(cdr3_nucseq) ==
            len(v_nucseq) + len(nseq) + len(j_nucseq) - ( v_trim + j_trim ))

    #d_info = ''
    n_vj_insert = 0
    n_vd_insert = 0
    n_dj_insert = 0
    d0_trim = 0
    d1_trim = 0
    best_d_id = 0

    if ab == 'A':
        n_vj_insert = len(nseq)

    elif ab == 'B':
        ## look for one of the d-gene segments
        max_overlap = 0
        for d_id, d_nucseq in all_trbd_nucseq[organism].items():
            if force_d_id and d_id != force_d_id: continue
            for start in range(len(d_nucseq)):
                for stop in range(start,len(d_nucseq)):
                    overlap_seq = d_nucseq[start:stop+1]
                    # string containment only works if cdr3_nucseq is all acgt
                    if ((not cdr3_nucseq_has_other and overlap_seq in nseq) or
                        (cdr3_nucseq_has_other and
                         logo_tools.degnucs_contain(nseq, overlap_seq))):
                        if len(overlap_seq)>max_overlap:
                            max_overlap = len(overlap_seq)
                            best_d_id = d_id
                            best_overlap_seq = overlap_seq
                            best_trim = (start,len(d_nucseq)-1-stop)

        if max_overlap: # found a bit of d, although it might be bogus (eg 1 nt)
            pos = (logo_tools.degnucs_index(nseq, best_overlap_seq)
                   if cdr3_nucseq_has_other else nseq.index(best_overlap_seq))
            for i in range(pos+num_matched_v,pos+num_matched_v+max_overlap):
                assert ncount[i] == 1
                ncount[i] = 0
                cdr3_nucseq_src[i] = 'D'
                nseq = nseq[:i-num_matched_v] + '+' + nseq[i+1-num_matched_v:]

            n_vd_insert = pos
            n_dj_insert = len(nseq) - pos - len(best_overlap_seq)
            d0_trim = best_trim[0]
            d1_trim = best_trim[1]

            expected_cdr3_nucseq_len = (
                len(v_nucseq) + n_vd_insert +
                len(all_trbd_nucseq[organism][best_d_id]) + n_dj_insert +
                len(j_nucseq) -
                ( v_trim + d0_trim + d1_trim + j_trim ) )
            assert len(cdr3_nucseq) == expected_cdr3_nucseq_len


        else:
            best_d_id = 0
            n_vd_insert = 0
            n_dj_insert = 0
            n_vj_insert = len(nseq) # BUGFIX? added 2021-07-12
            d0_trim = 0 # this is not quite right, since D gene was likely completely
            d1_trim = 0 # trimmed... so total trims value will not make sense


    if cdr3_protseq:
        assert 3*len(cdr3_protseq) == len(ncount)

        newseq = ''
        newseq_ncount = ''

        for i,a in enumerate(cdr3_protseq):
            nc = sum(ncount[3*i:3*i+3])
            newseq_ncount += repr(nc)
            if nc>1:
                newseq += a
            elif nc==1:
                newseq += a.lower()
            else:
                newseq += '-'

        ## output
        cdr3_protseq_masked = newseq[:]
        cdr3_protseq_new_nucleotide_countstring = newseq_ncount[:]
    else:## no protseq given (perhaps its an out of frame?)
        cdr3_protseq_masked = ''
        cdr3_protseq_new_nucleotide_countstring = ''

    new_nucseq = nseq[:]

    trims = ( v_trim, d0_trim, d1_trim, j_trim )
    inserts = ( best_d_id, n_vd_insert, n_dj_insert, n_vj_insert )

    ## new_nucseq spans the inserted nucleotide sequence and has '+'
    ##  for D-nucleotides
    if return_cdr3_nucseq_src:
        return (new_nucseq,
                cdr3_protseq_masked,
                cdr3_protseq_new_nucleotide_countstring,
                trims,
                inserts,
                cdr3_nucseq_src,
        )
    else:
        return (new_nucseq,
                cdr3_protseq_masked,
                cdr3_protseq_new_nucleotide_countstring,
                trims,
                inserts,
        )


def parse_tcr_junctions( organism, tcrs):
    '''
    Analyze the junction regions of all the tcrs. Return a pandas dataframe with the results, in the same order
    as tcrs

    tcrs = [(atcr1, btcr1),...] and either atcr or btcr can be None
    '''

    dfl = []

    for ii, (atcr, btcr) in enumerate(tcrs):
        if ii%50000==0:
            print('Parsing TCR junctions:', ii, len(tcrs))

        results = OrderedDict()

        if atcr is not None:
            va, ja, cdr3a, cdr3a_nucseq = atcr

            aresults = analyze_junction(organism, va, ja, cdr3a, cdr3a_nucseq,
                                        return_cdr3_nucseq_src=True)

            # trims = ( v_trim, d0_trim, d1_trim, j_trim )
            # inserts = ( best_d_id, n_vd_insert, n_dj_insert, n_vj_insert )

            _, cdr3a_protseq_masked, _, a_trims, a_inserts, cdr3a_nucseq_src = aresults

            assert a_trims[1]+a_trims[2] == 0
            assert a_inserts[1]+a_inserts[2] == 0

            a_indels = f'+{a_inserts[3]}-{sum(a_trims)}'
            results.update(OrderedDict(
                clone_index=ii,
                va=va,
                ja=ja,
                cdr3a=cdr3a,
                cdr3a_nucseq=cdr3a_nucseq,
                cdr3a_protseq_masked=cdr3a_protseq_masked,
                cdr3a_nucseq_src=''.join(cdr3a_nucseq_src),
                va_trim=a_trims[0],
                ja_trim=a_trims[3],
                a_insert=a_inserts[3],
                a_indels=a_indels,
            ))

        if btcr is not None:
            vb, jb, cdr3b, cdr3b_nucseq = btcr

            bresults = analyze_junction(organism, vb, jb, cdr3b, cdr3b_nucseq,
                                        return_cdr3_nucseq_src=True)

            # trims = ( v_trim, d0_trim, d1_trim, j_trim )
            # inserts = ( best_d_id, n_vd_insert, n_dj_insert, n_vj_insert )

            _, cdr3b_protseq_masked, _, b_trims, b_inserts, cdr3b_nucseq_src = bresults

            b_indels = f'+{sum(b_inserts[1:])}-{sum(b_trims)}'

            results.update(OrderedDict(
                clone_index=ii,
                vb=vb,
                jb=jb,
                cdr3b=cdr3b,
                cdr3b_nucseq=cdr3b_nucseq,
                cdr3b_protseq_masked=cdr3b_protseq_masked,
                cdr3b_nucseq_src=''.join(cdr3b_nucseq_src),
                vb_trim=b_trims[0],
                d0_trim=b_trims[1],
                d1_trim=b_trims[2],
                jb_trim=b_trims[3],
                vd_insert=b_inserts[1],
                dj_insert=b_inserts[2],
                vj_insert=b_inserts[3],
                b_indels=b_indels,
            ))
        dfl.append(results)

    return pd.DataFrame(dfl)


def vj_compatible(v_gene, j_gene, organism):
    if organism == 'human_ig': # enforce kappa or lambda
        assert (v_gene.startswith('IG') and
                j_gene.startswith('IG') and
                v_gene[2] in 'HKL' and
                j_gene[2] in 'HKL')
        return v_gene[2] == j_gene[2]
    else:
        return True

def resample_shuffled_tcr_chains(
        organism,
        num_samples,
        chain, # 'A' or 'B'
        junctions_df, # dataframe made by the above function
        preserve_vj_pairings = False,
        return_src_junction_indices = False,
):
    ''' returns list of (v_gene, j_gene, cdr3, cdr3_nucseq) for inputting into tcrdist calcs (e.g.)

    Does not impose a TRBJ1 --> TRBD1 restriction, but that could
    be added...

    '''
    assert chain in ['A', 'B']
    # need list of (v_gene, j_gene, cdr3_nucseq, breakpoints_pre_d, breakpoints_post_d)
    # breakpoints sets could contain negative numbers: that means read from the back
    #
    junctions = []
    for ind, l in enumerate(junctions_df.itertuples()):
        # where are the acceptable breakpoints?
        # dont allow breakpoints in the middle of V D or J
        # the breakpoint is the index we can use as in  nucseq = nucseq1[:breakpoint] + nucseq2[breakpoint:]
        breakpoints_pre_d = set()
        breakpoints_post_d = set()
        nucseq_src = l.cdr3a_nucseq_src if chain == 'A' else l.cdr3b_nucseq_src
        #
        pre_d = True
        post_d = 'D' not in nucseq_src
        for ii in range(1,len(nucseq_src)):
            ## ii refers to the breakpoint between position ii-1 and position ii
            ## ie, right before position ii
            a,b = nucseq_src[ii-1], nucseq_src[ii]
            if a=='D':
                post_d = True
                pre_d = False
            if a==b and a!='N':
                #bad
                pass
            else:
                # -1 would be ii==len(nucseq_src)-1
                for bp in [ii, ii-len(nucseq_src)]:
                    if post_d:
                        breakpoints_post_d.add(bp)
                    if pre_d:
                        breakpoints_pre_d.add(bp)
        if chain=='A':
            junctions.append( (l.va, l.ja, l.cdr3a_nucseq,
                               breakpoints_pre_d, breakpoints_post_d, ind))
        else:
            junctions.append( (l.vb, l.jb, l.cdr3b_nucseq,
                               breakpoints_pre_d, breakpoints_post_d, ind))

    if preserve_vj_pairings:
        # setup a mapping from v/j genes to junctions
        print('create gene --> junctions mapping')
        v_gene2junctions, j_gene2junctions = {}, {}
        def trim_allele(g):
            return g[:g.index('*')]

        for j in junctions:
            vg = trim_allele(j[0])
            jg = trim_allele(j[1])
            v_gene2junctions.setdefault(vg,[]).append(j)
            j_gene2junctions.setdefault(jg,[]).append(j)
        #debugging
        v_counts = Counter({x:len(y) for x,y in v_gene2junctions.items()})
        j_counts = Counter({x:len(y) for x,y in j_gene2junctions.items()})
        for v,count in v_counts.most_common():
            print(f'v_count: {count:5d} {count/len(junctions):.3f} {v}')
        for j,count in j_counts.most_common():
            print(f'j_count: {count:5d} {count/len(junctions):.3f} {j}')
        sys.stdout.flush()


    # repeat:
    # choose 2 random tcrs; are their breakpoints compatible?
    # if so, choose random compatible breakpoint, make frankentcr


    new_tcrs = []
    src_junction_indices = []
    attempts = 0
    successes = 0
    while len(new_tcrs) < num_samples:
        #print(len(new_tcrs))
        if preserve_vj_pairings:
            j = random.choice(junctions) # determines v and j genes
            t1 = random.choice(v_gene2junctions[trim_allele(j[0])])
            t2 = random.choice(j_gene2junctions[trim_allele(j[1])])
        else:
            t1 = random.choice(junctions)
            t2 = random.choice(junctions)
        nucseq1 = t1[2]
        nucseq2 = t2[2]
        if nucseq1 == nucseq2:
            continue

        if not vj_compatible(t1[0], t2[1], organism):
            continue

        inds = [3] if chain=='A' else [3,4]
        random.shuffle(inds)

        success = False
        for ind in inds:
            # ind=3: t1[ind] = breakpoints_pre_d
            # ind=4: t1[ind] = breakpoints_post_d
            shared = t1[ind] & t2[ind]
            if shared:
                bp = random.choice(list(shared))
                nucseq = nucseq1[: bp] + nucseq2[bp :]
                assert len(nucseq)%3==0
                if bp<0: # DEBUGGING
                    assert len(nucseq) == len(nucseq1)
                else:
                    assert len(nucseq) == len(nucseq2)
                # but is there a stop codon??
                cdr3 = translation.get_translation(nucseq)
                assert len(cdr3) == len(nucseq)//3
                if '*' not in cdr3:
                    success = True
                    t_new = ( t1[0], t2[1], cdr3, nucseq)
                    # print('t1', translation.get_translation(t1[2]), t1)
                    # print('t2', translation.get_translation(t2[2]), t2)
                    # print('tn', translation.get_translation(t_new[3]), t_new)
                    # print(bp, nucseq)
                    new_tcrs.append(t_new)
                    src_junction_indices.append((t1[-1], t2[-1], bp))
                    break
        #print('success:', success)
        attempts += 1
        successes += success
        if success and len(new_tcrs) >= num_samples:
            break
    # print(f'success_rate: {100.0*successes/attempts:.2f}')
    if return_src_junction_indices:
        return new_tcrs, src_junction_indices
    else:
        return new_tcrs



########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################

